from PySide6.QtWidgets import QApplication
from PySide6.QtGui import QFileOpenEvent
from PySide6.QtCore import QEvent
import sys
import os
import petab.v1 as petab

from .views import MainWindow
from .controllers import MainController
from .models import PEtabModel

from pathlib import Path


def find_example(path: Path) -> Path:
    while path.parent != path:
        if (path / "example").is_dir():
            return path / "example"
        path = path.parent
        
    raise FileNotFoundError("Could not find examples directory")


class PEtabGuiApp(QApplication):
    def __init__(self):
        super().__init__(sys.argv)

        # Load the stylesheet
        # self.apply_stylesheet()
        self.model = PEtabModel()
        self.view = MainWindow()
        self.controller = MainController(self.view, self.model)

        # hack to be discussed
        self.view.controller = self.controller

        if len(sys.argv) > 1 and os.path.isfile(sys.argv[1]):
            self.controller.open_file(sys.argv[1], mode="overwrite")

        self.view.show()

    def event(self, event):
        if event.type() == QEvent.FileOpen:
            openEvent = QFileOpenEvent(event)            
            self.controller.open_file(openEvent.file(), mode="overwrite")

        return super().event(event)

    def apply_stylesheet(self):
        """Load and apply the QSS stylesheet."""
        stylesheet_path = os.path.join(
            os.path.dirname(__file__), "stylesheet.css"
        )
        if os.path.exists(stylesheet_path):
            with open(stylesheet_path, "r") as f:
                self.setStyleSheet(f.read())
        else:
            print(f"Warning: Stylesheet '{stylesheet_path}' not found!")


def main():
    app = PEtabGuiApp()
    sys.exit(app.exec())

if __name__ == "__main__":
    main()



import pandas as pd
from PySide6.QtCore import (Qt, QAbstractTableModel, QModelIndex, Signal,
                            QSortFilterProxyModel, QMimeData)
from PySide6.QtGui import QColor, QBrush, QPalette
from PySide6.QtWidgets import QApplication
from ..C import COLUMNS
from ..utils import validate_value, create_empty_dataframe, is_invalid, \
    get_selected
from ..controllers.default_handler import DefaultHandlerModel
from ..settings_manager import settings_manager


class PandasTableModel(QAbstractTableModel):
    """Basic table model for a pandas DataFrame."""
    # Signals
    relevant_id_changed = Signal(str, str, str)  # new_id, old_id, type
    new_log_message = Signal(str, str)  # message, color
    cell_needs_validation = Signal(int, int)  # row, column
    something_changed = Signal(bool)
    inserted_row = Signal(QModelIndex)
    fill_defaults = Signal(QModelIndex)

    def __init__(self, data_frame, allowed_columns, table_type, parent=None):
        super().__init__(parent)
        self._allowed_columns = allowed_columns
        self.table_type = table_type
        self._invalid_cells = set()
        self.highlighted_cells = set()
        self._has_named_index = False
        if data_frame is None:
            data_frame = create_empty_dataframe(allowed_columns, table_type)
        self._data_frame = data_frame
        # add a view here, access is needed for selectionModels
        self.view = None
        # offset for row and column to get from the data_frame to the view
        self.row_index_offset = 0
        self.column_offset = 0
        # default values setup
        self.config = settings_manager.get_table_defaults(table_type)
        self.default_handler = DefaultHandlerModel(self, self.config)

    def rowCount(self, parent=QModelIndex()):
        return self._data_frame.shape[0] + 1  # empty row at the end

    def columnCount(self, parent=QModelIndex()):
        return self._data_frame.shape[1] + self.column_offset

    def data(self, index, role=Qt.DisplayRole):
        """Return the data at the given index and role for the View."""
        if not index.isValid():
            return None
        row, column = index.row(), index.column()
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if row == self._data_frame.shape[0]:
                if column == 0:
                    return f"New {self.table_type}"
                return ""
            if column == 0:
                value = self._data_frame.index[row]
                return str(value)
            value = self._data_frame.iloc[row, column - 1]
            if is_invalid(value):
                return ""
            return str(value)
        elif role == Qt.BackgroundRole:
            return self.determine_background_color(row, column)
        elif role == Qt.ForegroundRole:
            # Return yellow text if this cell is a match
            if (row, column) in self.highlighted_cells:
                return QApplication.palette().color(QPalette.HighlightedText)
            return QBrush(QColor(0, 0, 0))  # Default black text
        return None

    def flags(self, index):
        """Return whether cells are editable and selectable"""
        if not index.isValid():
            return Qt.ItemIsEnabled
        return Qt.ItemIsEditable | Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        """Return the header data for the given section, orientation"""
        if role != Qt.DisplayRole:
            return None
        # role == Qt.DisplayRole
        if orientation == Qt.Horizontal:
            if section == 0:
                return self._data_frame.index.name
            else:
                return self._data_frame.columns[section - 1]
        if orientation == Qt.Vertical:
            return str(section)
        return None

    def insertRows(self, position, rows, parent=QModelIndex()) -> bool:
        """
        Insert new rows at the end of the DataFrame in-place.
        This function always adds rows at the end.

        Parameters:
        -----------
        position: Ignored, as rows are always inserted at the end.
        rows: The number of rows to add.
        parent: Unused in this implementation.

        Returns:
        --------
        bool: True if rows were added successfully.
        """
        end_position = len(self._data_frame)
        self.beginInsertRows(
            QModelIndex(), end_position, end_position + rows - 1
        )

        # In-place row addition using loc
        for i in range(rows):
            # Append an empty row or row with default values using loc
            self._data_frame.loc[end_position + i] = \
                [""] * self._data_frame.shape[1]
        self.endInsertRows()
        return True

    def insertColumn(self, column_name: str):
        """
        Override insertColumn to always add the column at the right (end) of the table,
        and do so in-place on the DataFrame.
        """
        if not (
            column_name in self._allowed_columns or
            self.table_type == "condition"
        ):  # empty dict means all columns allowed
            self.new_log_message.emit(
                f"Column '{column_name}' not allowed in {self.table_type} table",
                "orange"
            )
        position = self._data_frame.shape[1]
        self.beginInsertColumns(QModelIndex(), position, position)
        column_type = \
        self._allowed_columns.get(column_name, {"type": "STRING"})["type"]
        default_value = "" if column_type == "STRING" else 0
        self._data_frame[column_name] = default_value
        self.endInsertColumns()

        return True

    def setData(self, index, value, role=Qt.EditRole):
        if not (index.isValid() and role == Qt.EditRole):
            return False

        if role != Qt.EditRole:
            return False

        if is_invalid(value) or value == "":
            value = None
        # check whether multiple rows but only one column is selected
        multi_row_change, selected = self.check_selection()
        if not multi_row_change:
            return self._set_data_single(index, value)
        # multiple rows but only one column is selected
        all_set = list()
        for index in selected:
            all_set.append(self._set_data_single(index, value))
        return all(all_set)

    def _set_data_single(self, index, value):
        """Set the data of a single cell."""
        col_setoff = 0
        if self._has_named_index:
            col_setoff = 1
        if index.row() == self._data_frame.shape[0]:
            # empty row at the end
            self.insertRows(index.row(), 1)
            self.fill_defaults.emit(index)
            # self.get_default_values(index)
            next_index = self.index(index.row(), 0)
            self.inserted_row.emit(next_index)
        if index.column() == 0 and self._has_named_index:
            return self.handle_named_index(index, value)
        row, column = index.row(), index.column()
        # Handling non-index (regular data) columns
        column_name = self._data_frame.columns[column - col_setoff]
        old_value = self._data_frame.iloc[row, column - col_setoff]
        # cast to numeric if necessary
        if not self._data_frame[column_name].dtype == "object":
            try:
                value = float(value)
            except ValueError:
                self.new_log_message.emit(
                    f"Column '{column_name}' expects a numeric value",
                    "red"
                )
                return False
        if value == old_value:
            return False

        if column_name == "observableId":
            self._data_frame.iloc[row, column - col_setoff] = value
            self.dataChanged.emit(index, index, [Qt.DisplayRole])
            self.relevant_id_changed.emit(value, old_value, "observable")
            self.cell_needs_validation.emit(row, column)
            self.something_changed.emit(True)
            return True
        if column_name in ["conditionId", "simulationConditionId",
                           "preequilibrationConditionId"]:
            self._data_frame.iloc[row, column - col_setoff] = value
            self.dataChanged.emit(index, index, [Qt.DisplayRole])
            self.relevant_id_changed.emit(value, old_value, "condition")
            self.cell_needs_validation.emit(row, column)
            self.something_changed.emit(True)
            return True

        # Validate data based on expected type
        expected_type = self._allowed_columns.get(column_name, None)
        if expected_type:
            expected_type = expected_type["type"]
            tried_value = value
            value, error_message = validate_value(
                value, expected_type
            )
            if error_message:
                self.new_log_message.emit(
                    f"Column '{column_name}' expects a value of "
                    f"type {expected_type}, but got '{tried_value}'",
                    "red"
                )
                return False
        # Set the new value
        self._data_frame.iloc[row, column - col_setoff] = value
        # Validate the row after setting data
        self.cell_needs_validation.emit(row, column)
        self.something_changed.emit(True)
        self.dataChanged.emit(index, index, [Qt.DisplayRole])

        return True

    def handle_named_index(self, index, value):
        """Handle the named index column."""
        pass

    def get_default_values(self, index):
        """Return the default values for a the row in a new index."""
        pass

    def replace_text(self, old_text: str, new_text: str):
        """Replace text in the table."""
        # find all occurences of old_text and sae indices
        mask = self._data_frame.eq(old_text)
        if mask.any().any():
            self._data_frame.replace(old_text, new_text, inplace=True)
            # Get first and last modified cell for efficient `dataChanged` emit
            changed_cells = mask.stack()[
                mask.stack()].index.tolist()  # Extract (row, col) pairs
            if changed_cells:
                first_row, first_col = changed_cells[0]
                last_row, last_col = changed_cells[-1]
                if self._has_named_index:
                    first_col += 1
                    last_col += 1
                top_left = self.index(first_row, first_col)
                bottom_right = self.index(last_row, last_col)
                self.dataChanged.emit(top_left, bottom_right, [Qt.DisplayRole])
        # also replace in the index
        if self._has_named_index and old_text in self._data_frame.index:
            self._data_frame.rename(index={old_text: new_text}, inplace=True)
            index_row = self._data_frame.index.get_loc(new_text)
            index_top_left = self.index(index_row, 0)
            index_bottom_right = self.index(index_row, 0)
            self.dataChanged.emit(
                index_top_left, index_bottom_right, [Qt.DisplayRole]
            )

    def get_df(self):
        """Return the DataFrame."""
        return self._data_frame

    def add_invalid_cell(self, row, column):
        """Add an invalid cell to the set."""
        # check that the index is valid
        if not self.index(row, column).isValid():
            return
        # return if it is the last row
        if row == self._data_frame.shape[0]:
            return
        # return if it is already invalid
        if (row, column) in self._invalid_cells:
            return
        self._invalid_cells.add((row, column))
        self.dataChanged.emit(
            self.index(row, column),
            self.index(row, column),
            [Qt.BackgroundRole]
        )

    def discard_invalid_cell(self, row, column):
        """Discard an invalid cell from the set."""
        self._invalid_cells.discard((row, column))
        self.dataChanged.emit(
            self.index(row, column),
            self.index(row, column),
            [Qt.BackgroundRole]
        )

    def update_invalid_cells(self, selected, mode: str = "rows"):
        """Edits the invalid cells when values are deleted."""
        if not selected:
            return
        old_invalid_cells = self._invalid_cells.copy()
        new_invalid_cells = set()
        sorted_to_del = sorted(selected)
        for a, b in old_invalid_cells:
            if mode == "rows":
                to_be_change = a
                not_changed = b
            elif mode == "columns":
                to_be_change = b
                not_changed = a
            if to_be_change in selected:
                continue
            smaller_count = sum(1 for x in sorted_to_del if x < to_be_change)
            new_val = to_be_change - smaller_count
            new_invalid_cells.add((new_val, not_changed))
        self._invalid_cells = new_invalid_cells

    def notify_data_color_change(self, row, column):
        """Notify the view to change the color of some cells"""
        self.dataChanged.emit(
            self.index(row, column),
            self.index(row, column),
            [Qt.BackgroundRole]
        )

    def get_value_from_column(self, column_name, row):
        """Retrieve the value from a specific column and row."""
        # if row is a new row return ""
        if row == self._data_frame.shape[0]:
            return ""
        if column_name in self._data_frame.columns:
            return self._data_frame.loc[row, column_name]
        if column_name == self._data_frame.index.name:
            return self._data_frame.index[row]
        return ""

    def return_column_index(self, column_name):
        """Return the index of a column. Defined in Subclasses"""
        pass

    def unique_values(self, column_name):
        """Return the unique values in a column."""
        if column_name in self._data_frame.columns:
            return list(self._data_frame[column_name].dropna().unique())
        if column_name == self._data_frame.index.name:
            return list(self._data_frame.index.dropna().unique())
        return []

    def delete_row(self, row):
        """Delete a row from the table."""
        self.beginRemoveRows(QModelIndex(), row, row)
        self._data_frame.drop(self._data_frame.index[row], inplace=True)
        self.endRemoveRows()

    def delete_column(self, column_index):
        """Delete a column from the DataFrame."""
        column_name = self._data_frame.columns[column_index - self.column_offset]
        self.beginRemoveColumns(QModelIndex(), column_index, column_index)
        self._data_frame.drop(columns=[column_name], inplace=True)
        self.endRemoveColumns()

    def clear_table(self):
        """Clear the table."""
        self.beginResetModel()
        self._data_frame.drop(self._data_frame.index, inplace=True)
        self.endResetModel()

    def check_selection(self):
        """Check if multiple rows but only one column is selected."""
        if self.view is None:
            return False
        selected = get_selected(self.view, mode="index")
        cols = set([index.column() for index in selected])
        rows = set([index.row() for index in selected])
        return len(rows) > 1 and len(cols) == 1, selected

    def reset_invalid_cells(self):
        """Reset the invalid cells and update their background color."""
        if not self._invalid_cells:
            return

        invalid_cells = list(self._invalid_cells)
        self._invalid_cells.clear()  # Clear invalid cells set

        for row, col in invalid_cells:
            index = self.index(row, col)
            self.dataChanged.emit(index, index, [Qt.BackgroundRole])

    def mimeData(self, rectangle, start_index):
        """Return the data to be copied to the clipboard.

        Parameters
        ----------
        rectangle: np.ndarray
            The rectangle of selected cells. Creates a minimum rectangle
            around all selected cells and is True if the cell is selected.
        start_index: (int, int)
            The start index of the selection. Used to determine the location
            of the copied data.
        """
        copied_data = ""
        for row in range(rectangle.shape[0]):
            for col in range(rectangle.shape[1]):
                if rectangle[row, col]:
                    copied_data += self.data(
                        self.index(start_index[0] + row, start_index[1] + col),
                        Qt.DisplayRole
                    )
                else:
                    copied_data += "SKIP"
                if col < rectangle.shape[1] - 1:
                    copied_data += "\t"
            copied_data += "\n"
        mime_data = QMimeData()
        mime_data.setText(copied_data.strip())
        return mime_data

    def setDataFromText(self, text, start_row, start_column):
        """Set the data from text."""
        # TODO: Does this need to be more flexible in the separator?
        lines = text.split("\n")
        self.maybe_add_rows(start_row, len(lines))
        for row_offset, line in enumerate(lines):
            values = line.split("\t")
            for col_offset, value in enumerate(values):
                if value == "SKIP":
                    continue
                self.setData(
                    self.index(
                        start_row + row_offset, start_column + col_offset
                    ),
                    value,
                    Qt.EditRole
                )

    def maybe_add_rows(self, start_row, n_rows):
        """Add rows if needed."""
        if start_row + n_rows > self._data_frame.shape[0]:
            self.insertRows(
                self._data_frame.shape[0],
                start_row + n_rows - self._data_frame.shape[0]
            )

    def determine_background_color(self, row, column):
        """Determine the background color of a cell.

        1. If it is the first column and last row, return light green.
        2. If it is an invalid cell, return red
        3. If it is an even row return light blue
        4. Otherwise return light green
        """
        if (row, column) == (self._data_frame.shape[0], 0):
            return QColor(144, 238, 144, 150)
        if (row, column) in self.highlighted_cells:
            return QApplication.palette().color(QPalette.Highlight)
        if (row, column) in self._invalid_cells:
            return QColor(255, 100, 100, 150)
        if row % 2 == 0:
            return QColor(144, 190, 109, 102)
        return QColor(177, 217, 231, 102)

    def allow_column_deletion(self, column: int) -> bool:
        """Checks whether the column can safely be deleted"""
        if column == 0 and self._has_named_index:
            return False, self._data_frame.index.name
        column_name = self._data_frame.columns[column-self.column_offset]
        if column_name not in self._allowed_columns.keys():
            return True, column_name
        return self._allowed_columns[column_name]["optional"], column_name

    def endResetModel(self):
        """Override endResetModel to reset the default handler."""
        super().endResetModel()
        self.config = settings_manager.get_table_defaults(self.table_type)
        self.default_handler = DefaultHandlerModel(self, self.config)


class IndexedPandasTableModel(PandasTableModel):
    """Table model for tables with named index."""
    condition_2be_renamed = Signal(str, str)  # Signal to mother controller

    def __init__(self, data_frame, allowed_columns, table_type, parent=None):
        super().__init__(
            data_frame=data_frame,
            allowed_columns=allowed_columns,
            table_type=table_type,
            parent=parent
        )
        self._has_named_index = True
        self.column_offset = 1

    def get_default_values(self, index):
        """Return the default values for a the row in a new index."""
        row = index.row()
        if isinstance(row, int):
            row = self._data_frame.index[row]
        columns_with_index = (
            [self._data_frame.index.name or "index"] +
            list(self._data_frame.columns)
        )
        for colname in columns_with_index:
            if colname == self._data_frame.index.name and not isinstance(row, int):
                continue
            if colname == self._data_frame.index.name and isinstance(row, int):
                default_value = self.default_handler.get_default(colname, row)
                if default_value == "":
                    default_value = f"{self.table_type}_{row}"
                self._data_frame.rename(
                    index={self._data_frame.index[row]: default_value},
                    inplace=True
                )
                row = default_value  # Update row to new index
                continue
            # if column is empty, fill with default value
            if self._data_frame.loc[row, colname] == "":
                default_value = self.default_handler.get_default(colname, row)
                self._data_frame.loc[row, colname] = default_value

    def handle_named_index(self, index, value):
        """Handle the named index column."""
        row, column = index.row(), index.column()
        old_value = self._data_frame.index[row]
        if value == old_value:
            return False
        if value in self._data_frame.index:
            self.new_log_message.emit(
                f"Duplicate index value '{value}'",
                "red"
            )
            return False
        try:
            self._data_frame.rename(index={old_value: value}, inplace=True)
            self.dataChanged.emit(index, index, [Qt.DisplayRole])
            self.relevant_id_changed.emit(value, old_value, self.table_type)
            self.cell_needs_validation.emit(row, 0)
            self.something_changed.emit(True)
            return True
        except Exception as e:
            self.new_log_message.emit(
                f"Error renaming index value '{old_value}' to '{value}': {e}",
                "red"
            )
            return False

    def return_column_index(self, column_name):
        """Return the index of a column."""
        if column_name in self._data_frame.columns:
            return self._data_frame.columns.get_loc(column_name) + 1
        if column_name == self._data_frame.index.name:
            return 0
        return -1


class MeasurementModel(PandasTableModel):
    """Table model for the measurement data."""
    possibly_new_condition = Signal(str)  # Signal for new condition
    possibly_new_observable = Signal(str)  # Signal for new observable

    def __init__(self, data_frame, parent=None):
        super().__init__(
            data_frame=data_frame,
            allowed_columns=COLUMNS["measurement"],
            table_type="measurement",
            parent=parent
        )

    def get_default_values(self, index):
        """Fill missing values in a row without modifying the index."""
        row = index.row()
        if isinstance(row, int):
            row_key = self._data_frame.index[row]
        else:
            row_key = row

        for colname in self._data_frame.columns:
            if self._data_frame.at[row_key, colname] == "":
                default_value = self.default_handler.get_default(colname,
                                                                 row_key)
                self._data_frame.at[row_key, colname] = default_value

    def data(self, index, role=Qt.DisplayRole):
        """Return the data at the given index and role for the View."""
        if not index.isValid():
            return None
        row, column = index.row(), index.column()
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if row == self._data_frame.shape[0]:
                if column == 0:
                    return f"New {self.table_type}"
                return ""
            value = self._data_frame.iloc[row, column]
            if is_invalid(value):
                return ""
            return str(value)
        elif role == Qt.BackgroundRole:
            return self.determine_background_color(row, column)
        elif role == Qt.ForegroundRole:
            # Return yellow text if this cell is a match
            if (row, column) in self.highlighted_cells:
                return QApplication.palette().color(QPalette.HighlightedText)
            return QBrush(QColor(0, 0, 0))  # Default black text
        return None

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        """Return the header data for the given section, orientation"""
        if role != Qt.DisplayRole:
            return None
        # role == Qt.DisplayRole
        if orientation == Qt.Horizontal:
            return self._data_frame.columns[section]
        if orientation == Qt.Vertical:
            return str(section)
        return None

    def fill_row(self, row_position: int, data: dict):
        """Fill a row with data.

        Parameters
        ----------
        row_position:
            The position of the row to fill.
        data:
            The data to fill the row with. Gets updated with default values.
        """
        data_to_add = {
            column_name: "" for column_name in self._data_frame.columns
        }
        # remove preequilibrationConditionId if not in columns
        if "preequilibrationConditionId" not in self._data_frame.columns:
            data.pop("preequilibrationConditionId", None)
        data_to_add.update(data)
        # Maybe add default values for missing columns
        self._data_frame.iloc[row_position] = data_to_add

    def return_column_index(self, column_name):
        """Return the index of a column."""
        if column_name in self._data_frame.columns:
            return self._data_frame.columns.get_loc(column_name)
        return -1


class ObservableModel(IndexedPandasTableModel):
    """Table model for the observable data."""

    def __init__(self, data_frame, parent=None):
        super().__init__(
            data_frame=data_frame,
            allowed_columns=COLUMNS["observable"],
            table_type="observable",
            parent=parent
        )

    def fill_row(self, row_position: int, data: dict):
        """Fill a row with data.

        Parameters
        ----------
        row_position:
            The position of the row to fill.
        data:
            The data to fill the row with. Gets updated with default values.
        """
        data_to_add = {
            column_name: "" for column_name in self._data_frame.columns
        }
        data_to_add.update(data)
        # Maybe add default values for missing columns?
        new_index = self._data_frame.index.tolist()
        index_name = self._data_frame.index.name
        new_index[row_position] = data_to_add.pop(
            "observableId"
        )
        self._data_frame.index = pd.Index(new_index, name=index_name)
        self._data_frame.iloc[row_position] = data_to_add
        # make a QModelIndex for the new row
        new_index = self.index(row_position, 0)
        self.fill_defaults.emit(new_index)


class ParameterModel(IndexedPandasTableModel):
    """Table model for the parameter data."""

    def __init__(self, data_frame, parent=None):
        super().__init__(
            data_frame=data_frame,
            allowed_columns=COLUMNS["parameter"],
            table_type="parameter",
            parent=parent
        )


class ConditionModel(IndexedPandasTableModel):
    """Table model for the condition data."""

    def __init__(self, data_frame, parent=None):
        super().__init__(
            data_frame=data_frame,
            allowed_columns=COLUMNS["condition"],
            table_type="condition",
            parent=parent
        )
        self._allowed_columns.pop("conditionId")

    def fill_row(self, row_position: int, data: dict):
        """Fill a row with data.

        Parameters
        ----------
        row_position:
            The position of the row to fill.
        data:
            The data to fill the row with. Gets updated with default values.
        """
        data_to_add = {
            column_name: "" for column_name in self._data_frame.columns
        }
        data_to_add.update(data)
        new_index = self._data_frame.index.tolist()
        index_name = self._data_frame.index.name
        new_index[row_position] = data_to_add.pop(
            "conditionId"
        )
        self._data_frame.index = pd.Index(new_index, name=index_name)
        self._data_frame.iloc[row_position] = data_to_add
        # make a QModelIndex for the new row
        new_index = self.index(row_position, 0)
        self.fill_defaults.emit(new_index)


class PandasTableFilterProxy(QSortFilterProxyModel):
    def __init__(self, model, parent=None):
        super().__init__(parent)
        self.source_model = model
        self.setSourceModel(model)

    def filterAcceptsRow(self, source_row, source_parent):
        """Custom filtering logic to apply global filtering across all columns."""
        source_model = self.sourceModel()

        # Always accept the last row (for "add new row")
        if source_row == source_model.rowCount() - 1:
            return True

        regex = self.filterRegularExpression()
        if regex.pattern() == "":
            return True

        for column in range(source_model.columnCount()):
            index = source_model.index(source_row, column, QModelIndex())
            data_str = str(source_model.data(index) or "")
            if regex.match(data_str).hasMatch():
                return True
        return False  # No match found

    def mimeData(self, rectangle, start_index):
        """Return the data to be copied to the clipboard."""
        return self.source_model.mimeData(rectangle, start_index)

    def setDataFromText(self, text, start_row, start_column):
        """Set the data from text."""
        return self.source_model.setDataFromText(text, start_row, start_column)

    @property
    def _invalid_cells(self):
        return self.source_model._invalid_cells



from PySide6.QtWidgets import (QDialog, QVBoxLayout, QHBoxLayout, QLabel, \
    QLineEdit, QPushButton, QCompleter, QCheckBox, QGridLayout, QTableView,
                               QWidget, QToolButton, QMenu)
from PySide6.QtGui import QSyntaxHighlighter, QTextCharFormat, QColor, QAction
from PySide6.QtCore import QObject, Signal, Qt
import re
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
import pandas as pd
import logging
from .C import ROW, COLUMN, INDEX
import antimony
import os
import math
import numpy as np
import qtawesome as qta


def _checkAntimonyReturnCode(code):
    """ Helper for checking the antimony response code.
    Raises Exception if error in antimony.

    :param code: antimony response
    :type code: int
    """
    if code < 0:
        raise Exception('Antimony: {}'.format(antimony.getLastError()))


def sbmlToAntimony(sbml):
    """ Convert SBML to antimony string.

    :param sbml: SBML string or file
    :type sbml: str | file
    :return: Antimony
    :rtype: str
    """
    antimony.clearPreviousLoads()
    antimony.freeAll()
    isfile = False
    try:
        isfile = os.path.isfile(sbml)
    except:
        pass
    if isfile:
        code = antimony.loadSBMLFile(sbml)
    else:
        code = antimony.loadSBMLString(str(sbml))
    _checkAntimonyReturnCode(code)
    return antimony.getAntimonyString(None)


def antimonyToSBML(ant):
    """ Convert Antimony to SBML string.

    :param ant: Antimony string or file
    :type ant: str | file
    :return: SBML
    :rtype: str
    """
    antimony.clearPreviousLoads()
    antimony.freeAll()
    try:
        isfile = os.path.isfile(ant)
    except ValueError:
        isfile = False
    if isfile:
        code = antimony.loadAntimonyFile(ant)
    else:
        code = antimony.loadAntimonyString(ant)
    _checkAntimonyReturnCode(code)
    mid = antimony.getMainModuleName()
    return antimony.getSBMLString(mid)


class ConditionInputDialog(QDialog):
    def __init__(self, condition_id, condition_columns, initial_values=None, error_key=None, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Add Condition")

        self.layout = QVBoxLayout(self)

        # Condition ID
        self.condition_id_layout = QHBoxLayout()
        self.condition_id_label = QLabel("Condition ID:", self)
        self.condition_id_input = QLineEdit(self)
        self.condition_id_input.setText(condition_id)
        self.condition_id_input.setReadOnly(True)
        self.condition_id_layout.addWidget(self.condition_id_label)
        self.condition_id_layout.addWidget(self.condition_id_input)
        self.layout.addLayout(self.condition_id_layout)

        # Dynamic fields for existing columns
        self.fields = {}
        for column in condition_columns:
            if column != "conditionId":  # Skip conditionId
                field_layout = QHBoxLayout()
                field_label = QLabel(f"{column}:", self)
                field_input = QLineEdit(self)
                if initial_values and column in initial_values:
                    field_input.setText(str(initial_values[column]))
                    if column == error_key:
                        field_input.setStyleSheet("background-color: red;")
                field_layout.addWidget(field_label)
                field_layout.addWidget(field_input)
                self.layout.addLayout(field_layout)
                self.fields[column] = field_input

        # Buttons
        self.buttons_layout = QHBoxLayout()
        self.ok_button = QPushButton("OK", self)
        self.cancel_button = QPushButton("Cancel", self)
        self.buttons_layout.addWidget(self.ok_button)
        self.buttons_layout.addWidget(self.cancel_button)
        self.layout.addLayout(self.buttons_layout)

        self.ok_button.clicked.connect(self.accept)
        self.cancel_button.clicked.connect(self.reject)

    def get_inputs(self):
        inputs = {column: field.text() for column, field in self.fields.items()}
        inputs["conditionId"] = self.condition_id_input.text()
        inputs["conditionName"] = inputs["conditionId"]
        return inputs


class MeasurementInputDialog(QDialog):
    def __init__(
        self,
        condition_ids = None,
        observable_ids = None,
        initial_values=None,
        error_key=None,
        parent=None
    ):
        super().__init__(parent)
        self.setWindowTitle("Add Measurement")

        self.layout = QVBoxLayout(self)

        # Observable ID
        self.observable_id_layout = QHBoxLayout()
        self.observable_id_label = QLabel("Observable ID:", self)
        self.observable_id_input = QLineEdit(self)
        if initial_values and "observableId" in initial_values:
            self.observable_id_input.setText(str(initial_values["observableId"]))
            if "observableId" == error_key:
                self.observable_id_input.setStyleSheet("background-color: red;")
        self.observable_id_layout.addWidget(self.observable_id_label)
        self.observable_id_layout.addWidget(self.observable_id_input)
        self.layout.addLayout(self.observable_id_layout)

        if observable_ids:
            # Auto-suggestion for Observable ID
            observable_completer = QCompleter(observable_ids, self)
            self.observable_id_input.setCompleter(observable_completer)

        # Measurement
        self.measurement_layout = QHBoxLayout()
        self.measurement_label = QLabel("Measurement:", self)
        self.measurement_input = QLineEdit(self)
        if initial_values and "measurement" in initial_values:
            self.measurement_input.setText(str(initial_values["measurement"]))
            if "measurement" == error_key:
                self.measurement_input.setStyleSheet("background-color: red;")
        self.measurement_layout.addWidget(self.measurement_label)
        self.measurement_layout.addWidget(self.measurement_input)
        self.layout.addLayout(self.measurement_layout)

        # Timepoints
        self.timepoints_layout = QHBoxLayout()
        self.timepoints_label = QLabel("Timepoints:", self)
        self.timepoints_input = QLineEdit(self)
        if initial_values and "time" in initial_values:
            self.timepoints_input.setText(str(initial_values["time"]))
            if "time" == error_key:
                self.timepoints_input.setStyleSheet("background-color: red;")
        self.timepoints_layout.addWidget(self.timepoints_label)
        self.timepoints_layout.addWidget(self.timepoints_input)
        self.layout.addLayout(self.timepoints_layout)

        # Condition ID
        self.condition_id_layout = QHBoxLayout()
        self.condition_id_label = QLabel("Condition ID:", self)
        self.condition_id_input = QLineEdit(self)
        if initial_values and "conditionId" in initial_values:
            self.condition_id_input.setText(str(initial_values["conditionId"]))
            if "conditionId" == error_key:
                self.condition_id_input.setStyleSheet("background-color: red;")
        elif condition_ids and len(condition_ids) == 1:
            self.condition_id_input.setText(condition_ids[0])
        self.condition_id_layout.addWidget(self.condition_id_label)
        self.condition_id_layout.addWidget(self.condition_id_input)
        self.layout.addLayout(self.condition_id_layout)

        if condition_ids:
            # Auto-suggestion for Condition ID
            condition_completer = QCompleter(condition_ids, self)
            self.condition_id_input.setCompleter(condition_completer)

        # Buttons
        self.buttons_layout = QHBoxLayout()
        self.ok_button = QPushButton("OK", self)
        self.cancel_button = QPushButton("Cancel", self)
        self.buttons_layout.addWidget(self.ok_button)
        self.buttons_layout.addWidget(self.cancel_button)
        self.layout.addLayout(self.buttons_layout)

        self.ok_button.clicked.connect(self.accept)
        self.cancel_button.clicked.connect(self.reject)

    def get_inputs(self):
        return (self.observable_id_input.text(),
                self.measurement_input.text(),
                self.timepoints_input.text(),
                self.condition_id_input.text())


class ObservableInputDialog(QDialog):
    def __init__(self, initial_values=None, error_key=None, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Add Observable")

        self.layout = QVBoxLayout(self)

        # Observable ID
        self.observable_id_layout = QHBoxLayout()
        self.observable_id_label = QLabel("Observable ID:", self)
        self.observable_id_input = QLineEdit(self)
        if initial_values and "observableId" in initial_values:
            self.observable_id_input.setText(str(initial_values["observableId"]))
            if "observableId" == error_key:
                self.observable_id_input.setStyleSheet("background-color: red;")
        self.observable_id_layout.addWidget(self.observable_id_label)
        self.observable_id_layout.addWidget(self.observable_id_input)
        self.layout.addLayout(self.observable_id_layout)

        # Observable Formula
        self.observable_formula_layout = QHBoxLayout()
        self.observable_formula_label = QLabel("Observable Formula:", self)
        self.observable_formula_input = QLineEdit(self)
        if initial_values and "observableFormula" in initial_values:
            self.observable_formula_input.setText(str(initial_values["observableFormula"]))
            if "observableFormula" == error_key:
                self.observable_formula_input.setStyleSheet("background-color: red;")
        self.observable_formula_layout.addWidget(self.observable_formula_label)
        self.observable_formula_layout.addWidget(self.observable_formula_input)
        self.layout.addLayout(self.observable_formula_layout)

        # Buttons
        self.buttons_layout = QHBoxLayout()
        self.ok_button = QPushButton("OK", self)
        self.cancel_button = QPushButton("Cancel", self)
        self.buttons_layout.addWidget(self.ok_button)
        self.buttons_layout.addWidget(self.cancel_button)
        self.layout.addLayout(self.buttons_layout)

        self.ok_button.clicked.connect(self.accept)
        self.cancel_button.clicked.connect(self.reject)

    def get_inputs(self):
        return self.observable_id_input.text(), self.observable_formula_input.text()


class ObservableFormulaInputDialog(QDialog):
    def __init__(self, observable_id, parent=None):
        super().__init__(parent)
        self.setWindowTitle(
            "You added a new observable! Please provide the formula."
        )

        self.layout = QVBoxLayout(self)

        # Observable ID
        self.observable_id_layout = QHBoxLayout()
        self.observable_id_label = QLabel("Observable ID:", self)
        self.observable_id_input = QLineEdit(self)
        self.observable_id_input.setText(observable_id)
        self.observable_id_input.setReadOnly(True)
        self.observable_id_layout.addWidget(self.observable_id_label)
        self.observable_id_layout.addWidget(self.observable_id_input)
        self.layout.addLayout(self.observable_id_layout)

        # Observable Formula
        self.observable_formula_layout = QHBoxLayout()
        self.observable_formula_label = QLabel("Observable Formula:", self)
        self.observable_formula_input = QLineEdit(self)
        self.observable_formula_layout.addWidget(self.observable_formula_label)
        self.observable_formula_layout.addWidget(self.observable_formula_input)
        self.layout.addLayout(self.observable_formula_layout)

        # Buttons
        self.buttons_layout = QHBoxLayout()
        self.ok_button = QPushButton("OK", self)
        self.cancel_button = QPushButton("Cancel", self)
        self.buttons_layout.addWidget(self.ok_button)
        self.buttons_layout.addWidget(self.cancel_button)
        self.layout.addLayout(self.buttons_layout)

        self.ok_button.clicked.connect(self.accept)
        self.cancel_button.clicked.connect(self.reject)

    def get_inputs(self):
        return (self.observable_id_input.text(),
                self.observable_formula_input.text())


class ParameterInputDialog(QDialog):
    def __init__(self, initial_values=None, error_key=None, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Add Parameter")

        self.layout = QVBoxLayout(self)

        # Parameter ID
        self.parameter_id_layout = QHBoxLayout()
        self.parameter_id_label = QLabel("Parameter ID:", self)
        self.parameter_id_input = QLineEdit(self)
        if initial_values and "parameterId" in initial_values:
            self.parameter_id_input.setText(str(initial_values["parameterId"]))
            if "parameterId" == error_key:
                self.parameter_id_input.setStyleSheet("background-color: red;")
        self.parameter_id_layout.addWidget(self.parameter_id_label)
        self.parameter_id_layout.addWidget(self.parameter_id_input)
        self.layout.addLayout(self.parameter_id_layout)

        # Nominal Value
        self.nominal_value_layout = QHBoxLayout()
        self.nominal_value_label = QLabel("Nominal Value (optional):", self)
        self.nominal_value_input = QLineEdit(self)
        if initial_values and "nominalValue" in initial_values:
            self.nominal_value_input.setText(str(initial_values["nominalValue"]))
            if "nominalValue" == error_key:
                self.nominal_value_input.setStyleSheet("background-color: red;")
        self.nominal_value_layout.addWidget(self.nominal_value_label)
        self.nominal_value_layout.addWidget(self.nominal_value_input)
        self.layout.addLayout(self.nominal_value_layout)

        # Buttons
        self.buttons_layout = QHBoxLayout()
        self.ok_button = QPushButton("OK", self)
        self.cancel_button = QPushButton("Cancel", self)
        self.buttons_layout.addWidget(self.ok_button)
        self.buttons_layout.addWidget(self.cancel_button)
        self.layout.addLayout(self.buttons_layout)

        self.ok_button.clicked.connect(self.accept)
        self.cancel_button.clicked.connect(self.reject)

    def get_inputs(self):
        return self.parameter_id_input.text(), self.nominal_value_input.text()


def set_dtypes(data_frame, columns, index_columns=None):
    dtype_mapping = {
        "STRING": str,
        "NUMERIC": float,
        "BOOLEAN": bool
    }
    for column, dtype in columns.items():
        if column in data_frame.columns:
            data_frame[column] = data_frame[column].astype(dtype_mapping[dtype])
    if index_columns:
        data_frame.set_index(index_columns, inplace=True)
    return data_frame


class FindReplaceDialog(QDialog):
    def __init__(self, parent=None, mode="petab", checkbox_states=None, controller=None):
        super().__init__(parent)
        self.controller = controller
        self.setWindowTitle("Find and Replace")
        self.mode = mode
        self.checkbox_states = checkbox_states or {}

        self.find_label = QLabel("Find:")
        self.find_input = QLineEdit()

        self.replace_label = QLabel("Replace:")
        self.replace_input = QLineEdit()

        self.find_button = QPushButton("Find")
        self.replace_button = QPushButton("Replace")
        self.close_button = QPushButton("Close")

        self.replace_button.clicked.connect(self.replace)
        self.close_button.clicked.connect(self.close)

        layout = QVBoxLayout()
        form_layout = QHBoxLayout()
        form_layout.addWidget(self.find_label)
        form_layout.addWidget(self.find_input)
        form_layout.addWidget(self.replace_label)
        form_layout.addWidget(self.replace_input)

        layout.addLayout(form_layout)

        checkbox_layout = QGridLayout()

        if self.mode == "petab":
            self.measurement_checkbox = QCheckBox("Measurement Table")
            self.observable_checkbox = QCheckBox("Observable Table")
            self.parameter_checkbox = QCheckBox("Parameter Table")
            self.condition_checkbox = QCheckBox("Condition Table")

            checkbox_layout.addWidget(self.measurement_checkbox, 0, 0)
            checkbox_layout.addWidget(self.observable_checkbox, 0, 1)
            checkbox_layout.addWidget(self.parameter_checkbox, 1, 0)
            checkbox_layout.addWidget(self.condition_checkbox, 1, 1)

            self.measurement_checkbox.setChecked(self.checkbox_states.get("measurement", False))
            self.observable_checkbox.setChecked(self.checkbox_states.get("observable", False))
            self.parameter_checkbox.setChecked(self.checkbox_states.get("parameter", False))
            self.condition_checkbox.setChecked(self.checkbox_states.get("condition", False))
        else:  # SBML mode
            self.sbml_checkbox = QCheckBox("SBML Text")
            self.antimony_checkbox = QCheckBox("Antimony Text")

            checkbox_layout.addWidget(self.sbml_checkbox, 0, 0)
            checkbox_layout.addWidget(self.antimony_checkbox, 0, 1)

            self.sbml_checkbox.setChecked(self.checkbox_states.get("sbml", False))
            self.antimony_checkbox.setChecked(self.checkbox_states.get("antimony", False))

        layout.addLayout(checkbox_layout)

        layout.addWidget(self.replace_button)
        layout.addWidget(self.close_button)
        self.setLayout(layout)

    def closeEvent(self, event):
        if self.mode == "petab":
            self.checkbox_states["measurement"] = self.measurement_checkbox.isChecked()
            self.checkbox_states["observable"] = self.observable_checkbox.isChecked()
            self.checkbox_states["parameter"] = self.parameter_checkbox.isChecked()
            self.checkbox_states["condition"] = self.condition_checkbox.isChecked()
        else:  # SBML mode
            self.checkbox_states["sbml"] = self.sbml_checkbox.isChecked()
            self.checkbox_states["antimony"] = self.antimony_checkbox.isChecked()
        super().closeEvent(event)

    def replace(self):
        find_text = self.find_input.text()
        replace_text = self.replace_input.text()

        if self.mode == "petab":
            if self.measurement_checkbox.isChecked():
                self.controller.measurement_controller.replace_text(find_text, replace_text)
            if self.observable_checkbox.isChecked():
                self.controller.observable_controller.replace_text(find_text, replace_text)
            if self.parameter_checkbox.isChecked():
                self.controller.parameter_controller.replace_text(find_text, replace_text)
            if self.condition_checkbox.isChecked():
                self.controller.condition_controller.replace_text(find_text, replace_text)
        else:  # SBML mode
            if self.sbml_checkbox.isChecked():
                sbml_text = self.parent().sbml_viewer.sbml_text_edit.toPlainText()
                sbml_text = sbml_text.replace(find_text, replace_text)
                self.parent().sbml_viewer.sbml_text_edit.setPlainText(sbml_text)

            if self.antimony_checkbox.isChecked():
                antimony_text = self.parent().sbml_viewer.antimony_text_edit.toPlainText()
                antimony_text = antimony_text.replace(find_text, replace_text)
                self.parent().sbml_viewer.antimony_text_edit.setPlainText(antimony_text)




class SyntaxHighlighter(QSyntaxHighlighter):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._rules = []

        # Define formats
        keyword_format = QTextCharFormat()
        keyword_format.setForeground(QColor("blue"))

        # Define regex patterns
        keywords = ["keyword1", "keyword2"]  # Replace with actual keywords
        keyword_pattern = r"\b(" + "|".join(keywords) + r")\b"
        self._rules.append((re.compile(keyword_pattern), keyword_format))

    def highlightBlock(self, text):
        for pattern, format in self._rules:
            for match in pattern.finditer(text):
                self.setFormat(match.start(), match.end() - match.start(), format)


def validate_value(value, expected_type):
    try:
        if expected_type == "STRING":
            value = str(value)
        elif expected_type == "NUMERIC":
            value = float(value)
        elif expected_type == "BOOLEAN":
            value = bool(value)
    except ValueError as e:
        return None, str(e)
    return value, None


class PlotWidget(FigureCanvas):

    def __init__(self, parent=None, width=5, height=4, dpi=100):
        fig = Figure(figsize=(width, height), dpi=dpi)
        self.axes = fig.add_subplot(111)
        super(PlotWidget, self).__init__(fig)


class SignalForwarder(QObject):
    """Forward signals from one object to another."""
    forwarded_signal = Signal()
    def __init__(self, original_signal):
        super().__init__()
        self.original_signal = original_signal
        self.original_signal.connect(self.forward_signal)

    def forward_signal(self, *args, **kwargs):
        """
        Capture any arguments from the original signal and forward them.
        """
        self.forwarded_signal.emit(*args, **kwargs)

    def connect_forwarded(self, slot):
        """Connect a slot to the forwarded signal."""
        self.forwarded_signal.connect(slot)


def create_empty_dataframe(column_dict: dict, table_type: str):
    columns = [col for col, props in column_dict.items() if not props["optional"]]
    dtypes = {
        col: props["type"] for col, props in column_dict.items() if not
        props["optional"]
    }
    df = pd.DataFrame(columns=columns).astype(dtypes)
    # set potential index columns
    if table_type == "observable":
        df.set_index("observableId", inplace=True)
    elif table_type == "parameter":
        df.set_index("parameterId", inplace=True)
    elif table_type == "condition":
        df.set_index("conditionId", inplace=True)
    return df


class CaptureLogHandler(logging.Handler):
    """A logging handler to capture log messages with levels."""
    def __init__(self):
        super().__init__()
        self.records = []  # Store full log records

    def emit(self, record):
        self.records.append(record)  # Save the entire LogRecord

    def get_formatted_messages(self):
        """Return formatted messages with levels."""
        return [
            f"{record.levelname}: {self.format(record)}" for record in self.records
        ]


def get_selected(table_view: QTableView, mode: str = ROW) -> list[int]:
    """
    Determines which rows are selected in a QTableView.

    Args:
        table_view (QTableView): The table view to check.

    Returns:
        list[int]: A list of selected row indices.
    """
    if not table_view or not isinstance(table_view, QTableView):
        return []
    if mode not in [ROW, COLUMN, INDEX]:
        return []

    selection_model = table_view.selectionModel()
    if not selection_model:
        return []
    selected_indexes = selection_model.selectedIndexes()
    if mode == INDEX:
        return selected_indexes
    if mode == COLUMN:
        selected_columns = set([index.column() for index in selected_indexes])
        return selected_columns
    if mode == ROW:
        selected_rows = set([index.row() for index in selected_indexes])
        return selected_rows
    return None


def get_selected_rectangles(table_view: QTableView) -> np.array:
    """Returns the selected cells in a rectangular view.

    The size of the rectangle is determined by Max_row - Min_row and
    Max_column - Min_column. The returned array is a boolean array with
    True values for selected cells.
    """
    selected = get_selected(table_view, mode=INDEX)
    if not selected:
        return None
    rows = [index.row() for index in selected]
    cols = [index.column() for index in selected]
    min_row, max_row = min(rows), max(rows)
    min_col, max_col = min(cols), max(cols)
    rect_start = (min_row, min_col)
    selected_rect = np.zeros(
        (max_row - min_row + 1, max_col - min_col + 1), dtype=bool
    )
    for index in selected:
        selected_rect[index.row() - min_row, index.column() - min_col] = True
    return selected_rect, rect_start


def process_file(filepath, logger):
    """
    Utility function to process a file based on its type and content.

    Args:
        filepath (str): Path to the file to process.
    """
    _, ext = os.path.splitext(filepath)
    ext = ext.lower()

    # Case 1: YAML files
    if ext in {".yaml", ".yml"}:
        return "yaml", None

    # Case 2: XML/SBML files
    if ext in {".xml", ".sbml"}:
        return "sbml", None

    # Case 3: CSV/TSV/TXT files
    if ext in {".csv", ".tsv", ".txt"}:
        # Determine separator by attempting to read the file with different delimiters
        separators = [",", "\t", ";"]
        separator = None
        header = None

        for sep in separators:
            # read the first line of the file
            try:
                with open(filepath, "r", encoding="utf-8") as file:
                    header = file.readline().strip().split(sep)
                if len(header) > 1:
                    separator = sep
                    break
            except Exception:
                continue

        if header is None:
            logger.log_message(
                f"Failed to read file: {filepath}. Perhaps unsupported "
                f"delimiter. Supported delimiters: {', '.join(separators)}",
                color="red"
            )
            return None, None

        # Case 3.2: Identify the table type based on header content
        if {"observableId", "measurement", "time"}.issubset(header):
            return "measurement", separator
        elif {"observableId", "observableFormula"}.issubset(header):
            return "observable", separator
        elif "parameterId" in header:
            return "parameter", separator
        elif "conditionId" in header or "\ufeffconditionId" in header:
            return "condition", separator
        else:
            logger.log_message(
                f"Unrecognized table type for file: {filepath}. Uploading as "
                f"data matrix.",
                color="orange"
            )
            return "data_matrix", separator
    logger.log_message(
        f"Unrecognized file type for file: {filepath}.",
        color="red"
    )
    return None, None


def is_invalid(value):
    """Check if a value is invalid."""
    if value is None:  # None values are invalid
        return True
    if isinstance(value, str):  # Strings can always be displayed
        return False
    try:
        return not math.isfinite(value)
    except TypeError:
        return True


class FindReplaceBar(QWidget):
    def __init__(self, controller, parent=None):
        super().__init__(parent)
        self.controller = controller
        self.controller_map = {
            "Observable Table": self.controller.observable_controller,
            "Condition Table": self.controller.condition_controller,
            "Parameter Table": self.controller.parameter_controller,
            "Measurement Table": self.controller.measurement_controller,
        }
        self.selected_controllers = set()
        self.only_search = False
        self.matches = None

        # 🔍 Find Input with options
        self.find_input = QLineEdit()
        self.find_input.setPlaceholderText("Find...")
        self.find_input.textChanged.connect(self.run_find)

        self.case_sensitive_button = QToolButton()
        self.case_sensitive_button.setIcon(qta.icon("mdi6.format-letter-case"))
        self.case_sensitive_button.setCheckable(True)
        self.case_sensitive_button.toggled.connect(self.run_find)

        self.word_match_button = QToolButton()
        self.word_match_button.setIcon(qta.icon("mdi6.alpha-w"))
        self.word_match_button.setCheckable(True)
        self.word_match_button.toggled.connect(self.run_find)

        self.regex_button = QToolButton()
        self.regex_button.setIcon(qta.icon("mdi6.regex"))
        self.regex_button.setCheckable(True)
        self.regex_button.toggled.connect(self.run_find)

        find_layout = QHBoxLayout()
        find_layout.addWidget(self.find_input)
        find_layout.addWidget(self.case_sensitive_button)
        find_layout.addWidget(self.word_match_button)
        find_layout.addWidget(self.regex_button)

        # 🔄 Replace Input
        self.replace_input = QLineEdit()
        self.replace_input.setPlaceholderText("Replace...")

        replace_layout = QHBoxLayout()
        replace_layout.addWidget(self.replace_input)

        # 🔘 Action Buttons (Navigation, Results, Replace, Close)
        self.prev_button = QPushButton()
        self.prev_button.setIcon(qta.icon("mdi6.arrow-up"))
        self.next_button = QPushButton()
        self.next_button.setIcon(qta.icon("mdi6.arrow-down"))
        self.prev_button.clicked.connect(self.find_previous)
        self.next_button.clicked.connect(self.find_next)

        self.results_label = QLabel("0 results")
        self.filter_button = QPushButton()
        self.filter_button.setIcon(qta.icon("mdi6.filter"))
        self.close_button = QPushButton()
        self.filter_button.clicked.connect(self.show_filter_menu)
        self.filter_menu = QMenu(self)  # Dropdown menu
        self.filter_actions = {}
        action = QAction("All", self.filter_menu)
        action.setCheckable(True)
        action.setChecked(True)
        action.triggered.connect(self.update_selected_controllers)
        self.filter_menu.addAction(action)
        self.filter_actions["All"] = action
        for table_name in self.controller_map.keys():
            action = QAction(table_name, self.filter_menu)
            action.setCheckable(True)
            action.triggered.connect(self.update_selected_controllers)
            self.filter_menu.addAction(action)
            self.filter_actions[table_name] = action
        self.close_button.setIcon(qta.icon("mdi6.close"))
        self.close_button.clicked.connect(self.hide)

        self.replace_button = QPushButton("Replace")
        self.replace_button.clicked.connect(self.replace_current_match)
        self.replace_all_button = QPushButton("Replace All")
        self.replace_all_button.clicked.connect(self.replace_all)

        find_controls_layout = QHBoxLayout()
        find_controls_layout.addWidget(self.results_label)
        find_controls_layout.addWidget(self.prev_button)
        find_controls_layout.addWidget(self.next_button)
        find_controls_layout.addWidget(self.filter_button)
        find_controls_layout.addWidget(self.close_button)

        replace_controls_layout = QHBoxLayout()
        replace_controls_layout.addWidget(self.replace_button)
        replace_controls_layout.addWidget(self.replace_all_button)

        # 🔹 Main Layout
        self.layout_main = QHBoxLayout()
        self.layout_edits = QVBoxLayout()
        self.layout_options = QVBoxLayout()

        self.layout_edits.addLayout(find_layout)
        self.layout_edits.addLayout(replace_layout)

        self.layout_options.addLayout(find_controls_layout)
        self.layout_options.addLayout(replace_controls_layout)

        self.layout_main.addLayout(self.layout_edits)
        self.layout_main.addLayout(self.layout_options)
        self.setLayout(self.layout_main)

    def run_find(self):
        """Triggered when the search text changes."""""
        search_text = self.find_input.text()
        case_sensitive = self.case_sensitive_button.isChecked()
        regex = self.regex_button.isChecked()
        whole_cell = self.word_match_button.isChecked()

        self.matches = []
        self.current_match_ind = -1

        for controller in [
            self.controller.observable_controller,
            self.controller.condition_controller,
            self.controller.parameter_controller,
            self.controller.measurement_controller,
        ]:
            matches = controller.find_text(search_text, case_sensitive, regex, whole_cell)
            self.matches.extend([(match[0], match[1], controller) for match in matches])  # Extend match with controller

        if self.matches:
            self.current_match_ind = 0
            self.focus_match(self.matches[self.current_match_ind])

        self.update_result_label()

    def find_next(self):
        """Move to the next match."""
        if not self.matches:
            return
        __, _, controller = self.matches[self.current_match_ind]
        controller.focus_match(None)
        self.current_match_ind = (self.current_match_ind + 1) % len(self.matches)
        row, col, controller = self.matches[self.current_match_ind]
        controller.focus_match((row, col), with_focus=True)
        self.update_result_label()

    def find_previous(self):
        """Move to the previous match."""
        if not self.matches:
            return
        __, _, controller = self.matches[self.current_match_ind]
        controller.focus_match(None)
        self.current_match_ind = (self.current_match_ind - 1) % len(
            self.matches)
        row, col, controller = self.matches[self.current_match_ind]
        controller.focus_match((row, col), with_focus=True)
        self.update_result_label()

    def update_result_label(self):
        """Update the result label dynamically."""
        match_count = len(self.matches)
        self.results_label.setText(
            f"{self.current_match_ind + 1}/{match_count}" if match_count > 0
            else "0 results"
        )

    def replace_current_match(self):
        """Replace the currently selected match and move to the next one."""
        if not self.matches or self.current_match_ind == -1:
            return

        replace_text = self.replace_input.text()
        if not replace_text:
            return

        row, col, controller = self.matches[self.current_match_ind]  # Extract controller from match

        controller.replace_text(
            row=row,
            col=col,
            replace_text=replace_text,
            search_text=self.find_input.text(),
            case_sensitive=self.case_sensitive_button.isChecked(),
            regex=self.regex_button.isChecked(),
        )
        # drop the current match and update the result label
        self.matches.pop(self.current_match_ind)
        self.update_result_label()
        match = self.matches[self.current_match_ind] if self.matches else None
        self.focus_match(match, with_focus=True)

    def replace_all(self):
        """Replace all matches with the given text."""
        if not self.matches:
            return

        replace_text = self.replace_input.text()
        search_text = self.find_input.text()
        case_sensitive = self.case_sensitive_button.isChecked()
        regex = self.regex_button.isChecked()

        controllers = set(match[2] for match in self.matches)  # Get unique controllers

        for controller in controllers:
            controller.replace_all(
                search_text, replace_text, case_sensitive, regex
            )
        # emit dataChanged emit for each match
        for row, col, controller in self.matches:
            controller.model.dataChanged.emit(
                controller.model.index(row, col),
                controller.model.index(row, col)
            )
            controller.cleanse_highlighted_cells()
        self.run_find()

    def focus_match(self, match, with_focus: bool = False):
        """Focus the match in the correct table."""
        if not match:
            return
        row, col, controller = match
        controller.focus_match((row, col), with_focus)

    def show_filter_menu(self):
        """Show the filter selection dropdown below the filter button."""
        self.filter_menu.exec_(self.filter_button.mapToGlobal(
            self.filter_button.rect().bottomLeft()))

    def update_selected_controllers(self):
        """Update which tables are included in the search based on selection."""
        if self.filter_actions["All"].isChecked():
            self.selected_controllers = set(self.controller_map.values())
        else:
            self.selected_controllers = {
                self.controller_map[table_name]
                for table_name, action in self.filter_actions.items() if
                action.isChecked() & (table_name != "All")
            }
        self.run_find()

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Escape:
            self.hide()
            event.accept()
        else:
            super().keyPressEvent(event)

    def hideEvent(self, event):
        """Reset highlights when the Find/Replace bar is hidden."""
        for controller in self.selected_controllers:
            controller.cleanse_highlighted_cells()
        super().hideEvent(event)

    def showEvent(self, event):
        """Reset highlights when the Find/Replace bar is shown."""
        # group matches by controller
        if not self.matches:
            super().showEvent(event)
            return
        for controller in [match[2] for match in self.matches]:
            matches = [
                (match[0], match[1]) for match in self.matches
                if match[2] == controller
            ]
            controller.highlight_text(matches)
        super().showEvent(event)

    def show_replace_parts(self, show: bool = False):
        """Toggle the visibility of the replace parts."""
        self.replace_input.setVisible(show)
        self.replace_button.setVisible(show)
        self.replace_all_button.setVisible(show)

    def toggle_find(self):
        """Toggle behaviour of the search bar."""
        if not self.isVisible():
            self.show()
            self.show_replace_parts(False)
            self.only_search = True
            return
        if not self.only_search:
            self.show_replace_parts(False)
            self.only_search = True
            return
        self.hide()

    def toggle_replace(self):
        """Toggle behaviour of the replace bar."""
        if not self.isVisible():
            self.show()
            self.show_replace_parts(True)
            self.only_search = False
            return
        if self.only_search:
            self.show_replace_parts(True)
            self.only_search = False
            return
        self.hide()



from functools import partial

from PySide6.QtWidgets import QMessageBox, QFileDialog, QLineEdit, QWidget, \
    QHBoxLayout, QToolButton, QTableView
from PySide6.QtGui import QAction, QDesktopServices
import zipfile
import tempfile
import os
from io import BytesIO
import logging
import yaml
import qtawesome as qta
from ..utils import FindReplaceDialog, CaptureLogHandler, process_file
from PySide6.QtCore import Qt, QUrl
from pathlib import Path
from ..models import PEtabModel
from .sbml_controller import SbmlController
from .table_controllers import MeasurementController, ObservableController, \
    ConditionController, ParameterController
from .logger_controller import LoggerController
from ..views import TaskBar
from .utils import prompt_overwrite_or_append, RecentFilesManager
from functools import partial
from ..settings_manager import SettingsDialog, settings_manager


class MainController:
    """Main controller class.

    Handles the communication between controllers. Handles general tasks.
    Mother controller to all other controllers. One controller to rule them
    all.
    """

    def __init__(self, view, model: PEtabModel):
        """Initialize the main controller.

        Parameters
        ----------
        view: MainWindow
            The main window.
        model: PEtabModel
            The PEtab model.
        """
        self.task_bar = None
        self.view = view
        self.model = model
        self.logger = LoggerController(view.logger_views)
        # CONTROLLERS
        self.measurement_controller = MeasurementController(
            self.view.measurement_dock,
            self.model.measurement,
            self.logger,
            self
        )
        self.observable_controller = ObservableController(
            self.view.observable_dock,
            self.model.observable,
            self.logger,
            self
        )
        self.parameter_controller = ParameterController(
            self.view.parameter_dock,
            self.model.parameter,
            self.logger,
            self
        )
        self.condition_controller = ConditionController(
            self.view.condition_dock,
            self.model.condition,
            self.logger,
            self
        )
        self.sbml_controller = SbmlController(
            self.view.sbml_viewer,
            self.model.sbml,
            self.logger,
            self
        )
        self.controllers = [
            self.measurement_controller,
            self.observable_controller,
            self.parameter_controller,
            self.condition_controller,
            self.sbml_controller
        ]
        # Recent Files
        self.recent_files_manager = RecentFilesManager(max_files=10)
        # Checkbox states for Find + Replace
        self.petab_checkbox_states = {
            "measurement": False,
            "observable": False,
            "parameter": False,
            "condition": False
        }
        self.sbml_checkbox_states = {
            "sbml": False,
            "antimony": False
        }
        self.unsaved_changes = False
        self.filter = QLineEdit()
        self.filter_active = {}  # Saves which tables the filter applies to
        self.actions = self.setup_actions()
        self.view.setup_toolbar(self.actions)

        self.setup_connections()
        self.setup_task_bar()
        self.setup_context_menu()

    def setup_context_menu(self):
        """Sets up context menus for the tables."""
        self.measurement_controller.setup_context_menu(self.actions)
        self.observable_controller.setup_context_menu(self.actions)
        self.parameter_controller.setup_context_menu(self.actions)
        self.condition_controller.setup_context_menu(self.actions)

    def setup_task_bar(self):
        """Create shortcuts for the main window."""
        self.view.task_bar = TaskBar(self.view, self.actions)
        self.task_bar = self.view.task_bar

    # CONNECTIONS
    def setup_connections(self):
        """Setup connections.

        Sets all connections that communicate from one different
        Models/Views/Controllers to another. Also sets general connections.
        """
        # Rename Observable
        self.observable_controller.observable_2be_renamed.connect(
            partial(
                self.measurement_controller.rename_value,
                column_names="observableId"
            )
        )
        # Rename Condition
        self.condition_controller.condition_2be_renamed.connect(
            partial(
                self.measurement_controller.rename_value,
                column_names = ["simulationConditionId",
                "preequilibrationConditionId"]
            )
        )
        # Add new condition or observable
        self.model.measurement.relevant_id_changed.connect(
            lambda x, y, z: self.observable_controller.maybe_add_observable(
                x, y) if z == "observable" else
            self.condition_controller.maybe_add_condition(
                x, y) if z == "condition" else None
        )
        # Maybe Move to a Plot Model
        self.view.measurement_dock.table_view.selectionModel().selectionChanged.connect(
            self.handle_selection_changed
        )
        self.model.measurement.dataChanged.connect(
            self.handle_data_changed
        )
        # Unsaved Changes
        self.model.measurement.something_changed.connect(
            self.unsaved_changes_change
        )
        self.model.observable.something_changed.connect(
            self.unsaved_changes_change
        )
        self.model.parameter.something_changed.connect(
            self.unsaved_changes_change
        )
        self.model.condition.something_changed.connect(
            self.unsaved_changes_change
        )
        self.model.sbml.something_changed.connect(
            self.unsaved_changes_change
        )
        # Visibility
        self.sync_visibility_with_actions()
        # Recent Files
        self.recent_files_manager.open_file.connect(
            partial(self.open_file, mode="overwrite")
        )
        # Settings logging
        settings_manager.new_log_message.connect(
            self.logger.log_message
        )

    def setup_actions(self):
        """Setup actions for the main controller."""
        actions = {"close": QAction(
            qta.icon("mdi6.close"),
            "&Close", self.view
        )}
        # Close
        actions["close"].setShortcut("Ctrl+Q")
        actions["close"].triggered.connect(self.view.close)
        # New File
        actions["new"] = QAction(
            qta.icon("mdi6.file-document"),
            "&New", self.view
        )
        actions["new"].setShortcut("Ctrl+N")
        actions["new"].triggered.connect(self.new_file)
        # Open File
        actions["open"] = QAction(
            qta.icon("mdi6.folder-open"),
            "&Open", self.view
        )
        actions["open"].setShortcut("Ctrl+O")
        actions["open"].triggered.connect(
            partial(self.open_file, mode="overwrite")
        )
        # Add File
        actions["add"] = QAction(
            qta.icon("mdi6.table-plus"),
            "Add", self.view
        )
        actions["add"].setShortcut("Ctrl+Shift+O")
        actions["add"].triggered.connect(
            partial(self.open_file, mode="append")
        )
        # Save
        actions["save"] = QAction(
            qta.icon("mdi6.content-save-all"),
            "&Save", self.view
        )
        actions["save"].setShortcut("Ctrl+S")
        actions["save"].triggered.connect(self.save_model)
        # Find + Replace
        actions["find"] = QAction(
            qta.icon("mdi6.magnify"),
            "Find", self.view
        )
        actions["find"].setShortcut("Ctrl+F")
        actions["find"].triggered.connect(self.find)
        actions["find+replace"] = QAction(
            qta.icon("mdi6.find-replace"),
            "Find/Replace", self.view
        )
        actions["find+replace"].setShortcut("Ctrl+R")
        actions["find+replace"].triggered.connect(self.replace)
        # Copy / Paste
        actions["copy"] = QAction(
            qta.icon("mdi6.content-copy"),
            "Copy", self.view
        )
        actions["copy"].setShortcut("Ctrl+C")
        actions["copy"].triggered.connect(self.copy_to_clipboard)
        actions["paste"] = QAction(
            qta.icon("mdi6.content-paste"),
            "Paste", self.view
        )
        actions["paste"].setShortcut("Ctrl+V")
        actions["paste"].triggered.connect(self.paste_from_clipboard)
        # add/delete row
        actions["add_row"] = QAction(
            qta.icon("mdi6.table-row-plus-after"),
            "Add Row", self.view
        )
        actions["add_row"].triggered.connect(self.add_row)
        actions["delete_row"] = QAction(
            qta.icon("mdi6.table-row-remove"),
            "Delete Row(s)", self.view
        )
        actions["delete_row"].triggered.connect(self.delete_rows)
        # add/delete column
        actions["add_column"] = QAction(
            qta.icon("mdi6.table-column-plus-after"),
            "Add Column", self.view
        )
        actions["add_column"].triggered.connect(self.add_column)
        actions["delete_column"] = QAction(
            qta.icon("mdi6.table-column-remove"),
            "Delete Column(s)", self.view
        )
        actions["delete_column"].triggered.connect(self.delete_column)
        # check petab model
        actions["check_petab"] = QAction(
            qta.icon("mdi6.checkbox-multiple-marked-circle-outline"),
            "Check PEtab", self.view
        )
        actions["check_petab"].triggered.connect(self.check_model)
        actions["reset_model"] = QAction(
            qta.icon("mdi6.restore"),
            "Reset SBML Model", self.view
        )
        actions["reset_model"].triggered.connect(
            self.sbml_controller.reset_to_original_model
        )
        # Recent Files
        actions["recent_files"] = self.recent_files_manager.tool_bar_menu

        # Filter widget
        filter_widget = QWidget()
        filter_layout = QHBoxLayout()
        filter_layout.setContentsMargins(0, 0, 0, 0)
        filter_widget.setLayout(filter_layout)
        self.filter_input = QLineEdit()
        self.filter_input.setPlaceholderText("Filter...")
        filter_layout.addWidget(self.filter_input)
        for table_n, table_name in zip(
            ["m", "p", "o", "c"],
            ["measurement", "parameter", "observable", "condition"]
        ):
            tool_button = QToolButton()
            icon = qta.icon(
                "mdi6.alpha-{}".format(table_n), "mdi6.filter",
                options=[
                    {'scale_factor': 1.5, 'offset': (-0.2, -0.2)},
                    {'off': 'mdi6.filter-off', 'offset': (0.3, 0.3)},
                ],
            )
            tool_button.setIcon(icon)
            tool_button.setCheckable(True)
            tool_button.setChecked(True)
            tool_button.setToolTip(f"Filter for {table_name} table")
            filter_layout.addWidget(tool_button)
            self.filter_active[table_name] = tool_button
            self.filter_active[table_name].toggled.connect(
                self.filter_table
            )
        actions["filter_widget"] = filter_widget
        self.filter_input.textChanged.connect(self.filter_table)

        # show/hide elements
        for element in ["measurement", "observable", "parameter", "condition"]:
            actions[f"show_{element}"] = QAction(
                f"{element.capitalize()} Table", self.view
            )
            actions[f"show_{element}"].setCheckable(True)
            actions[f"show_{element}"].setChecked(True)
        actions["show_logger"] = QAction(
            "Info", self.view
        )
        actions["show_logger"].setCheckable(True)
        actions["show_logger"].setChecked(True)
        actions["show_plot"] = QAction(
            "Data Plot", self.view
        )
        actions["show_plot"].setCheckable(True)
        actions["show_plot"].setChecked(True)
        # connect actions
        actions["reset_view"] = QAction(
            qta.icon("mdi6.view-grid-plus"),
            "Reset View", self.view
        )
        actions["reset_view"].triggered.connect(
            self.view.default_view
        )
        # Clear Log
        actions["clear_log"] = QAction(
            qta.icon("mdi6.delete"),
            "Clear Log", self.view
        )
        actions["clear_log"].triggered.connect(self.logger.clear_log)
        # Settings
        actions["settings"] = QAction(
            qta.icon("mdi6.cog"),
            "Settings", self.view
        )
        actions["settings"].triggered.connect(self.open_settings)

        # Opening the PEtab documentation
        actions["open_documentation"] = QAction(
            qta.icon("mdi6.web"),
            "View PEtab Documentation", self.view
        )
        actions["open_documentation"].triggered.connect(
            lambda: QDesktopServices.openUrl(QUrl(
                "https://petab.readthedocs.io/en/latest/v1/"
                "documentation_data_format.html"
            ))
        )

        return actions

    def sync_visibility_with_actions(self):
        """Sync dock visibility and QAction states in both directions."""

        dock_map = {
            "measurement": self.view.measurement_dock,
            "observable": self.view.observable_dock,
            "parameter": self.view.parameter_dock,
            "condition": self.view.condition_dock,
            "logger": self.view.logger_dock,
            "plot": self.view.plot_dock,
        }

        for key, dock in dock_map.items():
            action = self.actions[f"show_{key}"]

            # Initial sync: block signal to avoid triggering unwanted visibility changes
            was_blocked = action.blockSignals(True)
            action.setChecked(dock.isVisible())
            action.blockSignals(was_blocked)

            # Connect QAction ↔ DockWidget syncing
            action.toggled.connect(dock.setVisible)
            dock.visibilityChanged.connect(action.setChecked)

    def save_model(self):
        options = QFileDialog.Options()
        file_name, _ = QFileDialog.getSaveFileName(
            self.view,
            "Save Project",
            "",
            "Zip Files (*.zip)",
            options=options
        )
        if not file_name:
            return False
        if not file_name.endswith(".zip"):
            file_name += ".zip"

        # Create a temporary directory to save the model's files
        with tempfile.TemporaryDirectory() as temp_dir:
            self.model.save(temp_dir)

            # Create a bytes buffer to hold the zip file in memory
            buffer = BytesIO()
            with zipfile.ZipFile(buffer, 'w') as zip_file:
                # Add files to zip archive
                for root, _, files in os.walk(temp_dir):
                    for file in files:
                        file_path = os.path.join(root, file)
                        with open(file_path, 'rb') as f:
                            zip_file.writestr(file, f.read())
            with open(file_name, 'wb') as f:
                f.write(buffer.getvalue())

        QMessageBox.information(
            self.view, "Save Project",
            f"Project saved successfully to {file_name}"
        )
        return True

    def open_find_replace_dialog(self):
        current_tab = self.view.tab_widget.currentIndex()
        if current_tab == 0:
            # TODO: rewrite functionality in FindReplaceDialoge
            dialog = FindReplaceDialog(
                self.view, mode="petab",
                checkbox_states=self.petab_checkbox_states,
                controller=self
            )
        elif current_tab == 1:
            dialog = FindReplaceDialog(
                self.view, mode="sbml",
                checkbox_states=self.sbml_checkbox_states,
                controller=self
            )
        dialog.exec()

    def handle_selection_changed(self):
        # ??
        self.update_plot()

    def handle_data_changed(self, top_left, bottom_right, roles):
        # ??
        if not roles or Qt.DisplayRole in roles:
            self.update_plot()

    def update_plot(self):
        # ??
        selection_model = \
            self.view.measurement_dock.table_view.selectionModel()
        indexes = selection_model.selectedIndexes()
        if not indexes:
            return None

        selected_points = {}
        for index in indexes:
            if index.row() == self.model.measurement.get_df().shape[0]:
                continue
            row = index.row()
            observable_id = self.model.measurement._data_frame.iloc[row][
                "observableId"]
            if observable_id not in selected_points:
                selected_points[observable_id] = []
            selected_points[observable_id].append({
                "x": self.model.measurement._data_frame.iloc[row]["time"],
                "y": self.model.measurement._data_frame.iloc[row][
                    "measurement"]
            })
        if selected_points == {}:
            return None

        measurement_data = self.model.measurement._data_frame
        plot_data = {
            "all_data": [],
            "selected_points": selected_points
        }
        for observable_id in selected_points.keys():
            observable_data = measurement_data[
                measurement_data["observableId"] == observable_id]
            plot_data["all_data"].append({
                "observable_id": observable_id,
                "x": observable_data["time"].tolist(),
                "y": observable_data["measurement"].tolist()
            })

        self.view.plot_dock.update_visualization(plot_data)

    def open_file(self, file_path=None, mode=None):
        """Determines appropriate course of action for a given file.

        Course of action depends on file extension, separator and header
        structure. Opens the file in the appropriate controller.
        """
        if not file_path:
            file_path, _ = QFileDialog.getOpenFileName(
                self.view,
                "Open File",
                "",
                "All supported (*.yaml *.yml *.xml *.sbml *.tsv *.csv *.txt);;"
                "PEtab Problems (*.yaml *.yml);;SBML Files (*.xml *.sbml);;"
                "PEtab Tables or Data Matrix (*.tsv *.csv *.txt);;"
                "All files (*)"
            )
        if not file_path:
            return
        # handle file appropriately
        actionable, sep = process_file(file_path, self.logger)
        if actionable in ["yaml", "sbml"] and mode == "append":
            self.logger.log_message(
                f"Append mode is not supported for *.{actionable} files.",
                color="red"
            )
            return
        if not actionable:
            return
        if mode is None:
            if actionable in ["yaml", "sbml"]:
                mode = "overwrite"
            else:
                mode = prompt_overwrite_or_append(self)
        if mode is None:
            return
        self.recent_files_manager.add_file(file_path)
        self._open_file(actionable, file_path, sep, mode)

    def _open_file(self, actionable, file_path, sep, mode):
        """Overwrites the File in the appropriate controller.
        Actionable dictates which controller to use.
        """
        if actionable == "yaml":
            self.open_yaml_and_load_files(file_path)
        elif actionable == "sbml":
            self.sbml_controller.overwrite_sbml(file_path)
        elif actionable == "measurement":
            self.measurement_controller.open_table(
                file_path, sep, mode
            )
        elif actionable == "observable":
            self.observable_controller.open_table(
                file_path, sep, mode
            )
        elif actionable == "parameter":
            self.parameter_controller.open_table(
                file_path, sep, mode
            )
        elif actionable == "condition":
            self.condition_controller.open_table(
                file_path, sep, mode
            )
        elif actionable == "data_matrix":
            self.measurement_controller.process_data_matrix_file(
                file_path, mode, sep
            )

    def open_yaml_and_load_files(self, yaml_path=None, mode="overwrite"):
        """Open files from a YAML configuration.

        Opens a dialog to upload yaml file. Creates a PEtab problem and
        overwrites the current PEtab model with the new problem.
        """
        if not yaml_path:
            yaml_path, _ = QFileDialog.getOpenFileName(
                self.view,
                "Open YAML File",
                "",
                "YAML Files (*.yaml *.yml)"
            )
        if not yaml_path:
            return
        try:
            for controller in self.controllers:
                if controller == self.sbml_controller:
                    continue
                controller.release_completers()
            # Load the YAML content
            with open(yaml_path, 'r') as file:
                yaml_content = yaml.safe_load(file)

            # Resolve the directory of the YAML file to handle relative paths
            yaml_dir = Path(yaml_path).parent

            # Upload SBML model
            sbml_file_path = \
                yaml_dir / yaml_content['problems'][0]['sbml_files'][0]
            self.sbml_controller.overwrite_sbml(sbml_file_path)
            self.measurement_controller.open_table(
                yaml_dir / yaml_content['problems'][0]['measurement_files'][0]
            )
            self.observable_controller.open_table(
                yaml_dir / yaml_content['problems'][0]['observable_files'][0]
            )
            self.parameter_controller.open_table(
                yaml_dir / yaml_content['parameter_file']
            )
            self.condition_controller.open_table(
                yaml_dir / yaml_content['problems'][0]['condition_files'][0]
            )
            self.logger.log_message(
                "All files opened successfully from the YAML configuration.",
                color="green"
            )
            self.check_model()
            # rerun the completers
            for controller in self.controllers:
                if controller == self.sbml_controller:
                    continue
                controller.setup_completers()
            self.unsaved_changes = False

        except Exception as e:
            self.logger.log_message(
                f"Failed to open files from YAML: {str(e)}", color="red"
            )

    def new_file(self):
        """Empty all tables. In case of unsaved changes, ask to save."""
        if self.unsaved_changes:
            reply = QMessageBox.question(
                self.view, "Unsaved Changes",
                "You have unsaved changes. Do you want to save them?",
                QMessageBox.Save | QMessageBox.Discard | QMessageBox.Cancel,
                QMessageBox.Save
            )
            if reply == QMessageBox.Save:
                self.save_model()
        for controller in [
            self.measurement_controller,
            self.observable_controller,
            self.parameter_controller,
            self.condition_controller
        ]:
            controller.clear_table()

    def check_model(self):
        """Check the consistency of the model. And log the results."""
        capture_handler = CaptureLogHandler()
        logger = logging.getLogger("petab.v1.lint")  # Target the specific
        # logger
        logger.addHandler(capture_handler)

        try:
            # Run the consistency check
            failed = self.model.test_consistency()

            # Process captured logs
            if capture_handler.records:
                captured_output = "<br>&nbsp;&nbsp;&nbsp;&nbsp;".join(
                    capture_handler.get_formatted_messages()
                )
                self.logger.log_message(
                    f"Captured petab lint logs:<br>"
                    f"&nbsp;&nbsp;&nbsp;&nbsp;{captured_output}",
                    color="purple"
                )

            # Log the consistency check result
            if not failed:
                self.logger.log_message("Model is consistent.", color="green")
                for model in self.model.pandas_models.values():
                    model.reset_invalid_cells()
            else:
                self.logger.log_message("Model is inconsistent.", color="red")
        finally:
            # Always remove the capture handler
            logger.removeHandler(capture_handler)

    def unsaved_changes_change(self, unsaved_changes: bool):
        self.unsaved_changes = unsaved_changes
        if unsaved_changes:
            self.view.setWindowTitle("PEtab Editor - Unsaved Changes")
        else:
            self.view.setWindowTitle("PEtab Editor")

    def maybe_close(self):
        if not self.unsaved_changes:
            self.view.allow_close = True
            return
        reply = QMessageBox.question(
            self.view, "Unsaved Changes",
            "You have unsaved changes. Do you want to save them?",
            QMessageBox.Save | QMessageBox.Discard | QMessageBox.Cancel,
            QMessageBox.Save
        )
        if reply == QMessageBox.Save:
            saved = self.save_model()
            self.view.allow_close = saved
        elif reply == QMessageBox.Discard:
            self.view.allow_close = True
        else:
            self.view.allow_close = False

    def active_widget(self):
        active_widget = self.view.tab_widget.currentWidget()
        if active_widget == self.view.data_tab:
            active_widget = self.view.data_tab.focusWidget()
        if active_widget and isinstance(active_widget, QTableView):
            return active_widget
        return None

    def active_controller(self):
        active_widget = self.active_widget()
        if active_widget == self.view.measurement_dock.table_view:
            return self.measurement_controller
        if active_widget == self.view.observable_dock.table_view:
            return self.observable_controller
        if active_widget == self.view.parameter_dock.table_view:
            return self.parameter_controller
        if active_widget == self.view.condition_dock.table_view:
            return self.condition_controller
        print("No active controller found")
        return None

    def delete_rows(self):
        controller = self.active_controller()
        if controller:
            controller.delete_row()

    def add_row(self):
        controller = self.active_controller()
        if controller:
            controller.add_row()

    def add_column(self):
        controller = self.active_controller()
        if controller:
            controller.add_column()

    def delete_column(self):
        controller = self.active_controller()
        if controller:
            controller.delete_column()

    def filter_table(self):
        """Filter the currently activated tables"""
        filter_text = self.filter_input.text()
        for table_name, tool_button in self.filter_active.items():
            if tool_button.isChecked():
                controller = getattr(self, f"{table_name}_controller")
                controller.filter_table(filter_text)
            else:
                controller = getattr(self, f"{table_name}_controller")
                controller.remove_filter()

    def copy_to_clipboard(self):
        controller = self.active_controller()
        if controller:
            controller.copy_to_clipboard()

    def paste_from_clipboard(self):
        controller = self.active_controller()
        if controller:
            controller.paste_from_clipboard()

    def open_settings(self):
        """Opens the settings Dialogue."""
        # retrieve all current columns from the tables
        table_columns = {
            "observable": self.observable_controller.get_columns(),
            "parameter": self.parameter_controller.get_columns(),
            "measurement": self.measurement_controller.get_columns(),
            "condition": self.condition_controller.get_columns(),
        }
        settings_dialog = SettingsDialog(table_columns, self.view)
        settings_dialog.exec()

    def set_docks_visible(self):
        """Handles Visibility of docks."""
        pass

    def find(self):
        """Create a find replace bar if it is non existent."""
        if self.view.find_replace_bar is None:
            self.view.create_find_replace_bar()
        self.view.toggle_find()

    def replace(self):
        """Create a find replace bar if it is non existent."""
        if self.view.find_replace_bar is None:
            self.view.create_find_replace_bar()
        self.view.toggle_replace()



from PySide6.QtWidgets import (QDockWidget, QHeaderView, QTableView,
                               QCompleter, QLineEdit, QStyledItemDelegate,
                               QComboBox)
from PySide6.QtCore import Qt, QPropertyAnimation, QRect
from PySide6.QtGui import QGuiApplication, QColor

from ..utils import get_selected_rectangles
from .context_menu_mananger import ContextMenuManager
import re
import pandas as pd


class TableViewer(QDockWidget):
    def __init__(self, title, parent=None):
        super().__init__(title, parent)
        self.title = title
        self.setObjectName(title)
        self.setAllowedAreas(
            Qt.AllDockWidgetAreas
        )
        # Create the QTableView for the table content
        self.table_view = CustomTableView()
        self.setWidget(self.table_view)
        # Dictionary to store column-specific completers
        self.completers = {}
        self.table_view.setAlternatingRowColors(True)

    def copy_to_clipboard(self):
        selected_rect, rect_start = get_selected_rectangles(
            self.table_view
        )
        if selected_rect.any():
            mime_data = self.table_view.model().mimeData(
                selected_rect, rect_start
            )
            clipboard = QGuiApplication.clipboard()
            clipboard.setMimeData(mime_data)

    def paste_from_clipboard(self):
        clipboard = QGuiApplication.clipboard()
        text = clipboard.text()
        if not text:
            return
        start_index = self.table_view.selectionModel().currentIndex()
        if not start_index.isValid():
            return
        model = self.table_view.model()
        row_start, col_start = start_index.row(), start_index.column()
        # identify which invalid cells are being pasted into
        pasted_data = [line.split("\t") for line in text.split("\n") if
                       line.strip()]
        num_rows = len(pasted_data)
        num_cols = max([len(line) for line in pasted_data])
        overridden_cells = {
            (row_start + r, col_start + c)
            for r in range(num_rows)
            for c in range(num_cols)
            if model.index(row_start + r, col_start + c).isValid()
        }
        invalid_overridden_cells = overridden_cells.intersection(
            model._invalid_cells
        )
        if invalid_overridden_cells:
            for row_invalid, col_invalid in invalid_overridden_cells:
                model.discard_invalid_cell(row_invalid, col_invalid)

        model.setDataFromText(
            text, start_index.row(),
            start_index.column()
        )


class ComboBoxDelegate(QStyledItemDelegate):
    def __init__(self, options, parent=None):
        super().__init__(parent)
        self.options = options

    def createEditor(self, parent, option, index):
        # Create a QComboBox for inline editing
        editor = QComboBox(parent)
        editor.addItems(self.options)
        return editor


class SingleSuggestionDelegate(QStyledItemDelegate):
    """Suggest a single option based the current row and the value in
    `column_name`."""

    def __init__(self, model, suggestions_column, afix=None, parent=None):
        super().__init__(parent)
        self.model = model  # The main model to retrieve data from
        self.suggestions_column = suggestions_column
        self.afix = afix

    def createEditor(self, parent, option, index):
        # Create a QLineEdit for inline editing
        editor = QLineEdit(parent)

        # Get the conditionId of the current row
        row = index.row()
        suggestion = self.model.get_value_from_column(
            self.suggestions_column, row
        )
        if self.afix:
            suggestion = self.afix + suggestion

        # Set up the completer with a single suggestion
        completer = QCompleter([suggestion], parent)
        completer.setCompletionMode(QCompleter.InlineCompletion)
        editor.setCompleter(completer)

        return editor


class ColumnSuggestionDelegate(QStyledItemDelegate):
    """Suggest options based on all unique values in the specified column."""
    def __init__(
        self,
        model,
        suggestions_column,
        suggestion_mode=QCompleter.PopupCompletion,
        parent=None
    ):
        super().__init__(parent)
        self.model = model  # The main model to retrieve data from
        self.suggestions_column = suggestions_column
        self.suggestion_mode = suggestion_mode

    def createEditor(self, parent, option, index):
        # Create a QLineEdit for inline editing
        editor = QLineEdit(parent)

        # Get unique suggestions from the specified column
        suggestions = self.model.unique_values(self.suggestions_column)

        # Set up the completer with the unique values
        completer = QCompleter(suggestions, parent)
        completer.setCompletionMode(self.suggestion_mode)
        editor.setCompleter(completer)

        return editor


class ParameterIdSuggestionDelegate(QStyledItemDelegate):
    """Suggest options based on all unique values in the specified column."""
    def __init__(self, par_model, sbml_model, parent=None):
        super().__init__(parent)
        self.par_model = par_model
        self.sbml_model = sbml_model  # The main model to retrieve data from

    def createEditor(self, parent, option, index):
        # Create a QLineEdit for inline editing
        editor = QLineEdit(parent)

        # Get unique suggestions from the specified column
        curr_model = self.sbml_model.get_current_sbml_model()
        suggestions = None
        if curr_model:  # only if model is valid
            suggestions = curr_model.get_valid_parameters_for_parameter_table()
            # substract the current parameter ids except for the current row
            row = index.row()
            selected_parameter_id = self.par_model.get_value_from_column(
                'parameterId', row
            )
            current_parameter_ids = self.par_model.get_df().index.tolist()
            if selected_parameter_id in current_parameter_ids:
                current_parameter_ids.remove(selected_parameter_id)
            suggestions = list(set(suggestions) - set(current_parameter_ids))

        # Set up the completer with the unique values
        completer = QCompleter(suggestions, parent)
        completer.setCompletionMode(QCompleter.PopupCompletion)
        editor.setCompleter(completer)

        return editor


class CustomTableView(QTableView):
    """Custom Table View to Handle Copy Paste events, resizing policies etc."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setSizeAdjustPolicy(QTableView.AdjustToContents)
        self.horizontalHeader().setSectionResizeMode(
            QHeaderView.ResizeToContents
        )
        self.horizontalHeader().setStretchLastSection(
            False
        )  # Prevent last column from stretching

        self.horizontalHeader().sectionDoubleClicked.connect(
            self.autofit_column
        )

    def setup_context_menu(self, actions):
        """Setup the context menu for the table view."""
        self.context_menu_manager = ContextMenuManager(
            actions, self, self.parent
        )
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(
            self.context_menu_manager.create_context_menu
        )

    def setModel(self, model):
        """Ensures selection model exists before connecting signals"""
        super().setModel(model)
        if self.selectionModel():
            self.selectionModel().currentColumnChanged.connect(self.highlight_active_column)

    def reset_column_sizes(self):
        """Resets column sizes with refinements"""
        header = self.horizontalHeader()
        total_width = self.viewport().width()
        max_width = total_width // 4  # 1/4th of total table width

        header.setSectionResizeMode(QHeaderView.ResizeToContents)
        self.resizeColumnsToContents()
        header.setSectionResizeMode(QHeaderView.Interactive)

        # Enforce max width but allow expanding into empty neighbors
        for col in range(self.model().columnCount()):
            optimal_width = self.columnWidth(col)
            if optimal_width > max_width:
                self.setColumnWidth(col, max_width)
            else:
                self.setColumnWidth(col, optimal_width)

        # self.adjust_for_empty_neighbors()
        self.collapse_empty_columns()
        self.updateGeometry()

    def adjust_for_empty_neighbors(self):
        """Expands column if adjacent columns are empty"""
        model = self.model()
        for col in range(model.columnCount()):
            if self.columnWidth(col) == self.viewport().width() // 4:  # If maxed out
                next_col = col + 1
                if next_col < model.columnCount():
                    if all(model.index(row, next_col).data() in [None, ""] for row in range(model.rowCount())):
                        new_width = self.columnWidth(
                            col) + self.columnWidth(next_col)
                        self.setColumnWidth(col, new_width)
                        self.setColumnWidth(next_col, 0)  # Hide empty column

    def collapse_empty_columns(self):
        """Collapses columns that only contain empty values"""
        model = self.model()
        for col in range(model.columnCount()):
            if all(model.index(row, col).data() in [None, "", " "] for row in
                   range(model.rowCount())):
                self.setColumnWidth(col, 10)  # Minimal width

    def autofit_column(self, col):
        """Expands column width on double-click"""
        self.horizontalHeader().setSectionResizeMode(col,
                                                     QHeaderView.ResizeToContents)
        self.resizeColumnToContents(col)
        self.horizontalHeader().setSectionResizeMode(col,
                                                     QHeaderView.Interactive)

    def highlight_active_column(self, index):
        """Highlights the active column"""
        for row in range(self.model().rowCount()):
            self.model().setData(self.model().index(row, index.column()),
                                 QColor("#cce6ff"), Qt.BackgroundRole)

    def animate_column_resize(self, col, new_width):
        """Smoothly animates column resizing"""
        anim = QPropertyAnimation(self, b"geometry")
        anim.setDuration(200)
        anim.setStartValue(QRect(self.columnViewportPosition(col), 0,
                                 self.columnWidth(col), self.height()))
        anim.setEndValue(
            QRect(self.columnViewportPosition(col), 0, new_width,
                  self.height()))
        anim.start()



from PySide6.QtWidgets import QMessageBox, QMenu
from PySide6.QtCore import QObject, Signal
from PySide6.QtGui import QAction
from collections import Counter
from pathlib import Path

from ..settings_manager import settings_manager


def prompt_overwrite_or_append(controller):
    """Prompt user to choose between overwriting or appending the file."""
    msg_box = QMessageBox(controller.view)
    msg_box.setWindowTitle("Open File Options")
    msg_box.setText("Do you want to overwrite the current data or append to it?")
    overwrite_button = msg_box.addButton("Overwrite", QMessageBox.AcceptRole)
    append_button = msg_box.addButton("Append", QMessageBox.AcceptRole)
    cancel_button = msg_box.addButton("Cancel", QMessageBox.RejectRole)

    msg_box.exec()

    if msg_box.clickedButton() == cancel_button:
        return None
    elif msg_box.clickedButton() == overwrite_button:
        return "overwrite"
    elif msg_box.clickedButton() == append_button:
        return "append"


class RecentFilesManager(QObject):
    """Manage a list of recent files."""
    open_file = Signal(str)  # Signal to open a file

    def __init__(self, max_files=10):
        super().__init__()
        self.max_files = max_files
        self.recent_files = self.load_recent_files()
        self.tool_bar_menu = QMenu("Recent Files")
        self.update_tool_bar_menu()

    def add_file(self, file_path):
        """Add a file to the recent files list."""
        if file_path in self.recent_files:
            self.recent_files.remove(file_path)
        self.recent_files.insert(0, file_path)
        self.recent_files = self.recent_files[:self.max_files]
        self.save_recent_files()
        self.update_tool_bar_menu()

    @staticmethod
    def load_recent_files():
        """Load recent files from settings."""
        return settings_manager.get_value("recent_files", [])

    def save_recent_files(self):
        """Save recent files to settings."""
        settings_manager.set_value("recent_files", self.recent_files)

    def update_tool_bar_menu(self):
        """Update the recent files menu."""
        self.tool_bar_menu.clear()

        # Generate shortened names
        def short_name(path):
            p = Path(path)
            if p.parent.name:
                return f"{p.parent.name}/{p.name}"
            return p.name

        short_paths = [short_name(f) for f in self.recent_files]
        counts = Counter(short_paths)

        for full_path, short in zip(self.recent_files, short_paths):
            display = full_path if counts[short] > 1 else short
            action = QAction(display, self.tool_bar_menu)
            action.triggered.connect(lambda _, p=full_path: self.open_file.emit(p))
            self.tool_bar_menu.addAction(action)
        self.tool_bar_menu.addSeparator()
        clear_action = QAction("Clear Recent Files", self.tool_bar_menu)
        clear_action.triggered.connect(self.clear_recent_files)

    def clear_recent_files(self):
        """Clear the recent files list."""
        self.recent_files = []
        self.save_recent_files()
        self.update_tool_bar_menu()



"""The Default Handlers for the GUI."""
import pandas as pd
import numpy as np
import copy

from collections import Counter
from ..C import (COPY_FROM, USE_DEFAULT, NO_DEFAULT, MIN_COLUMN, MAX_COLUMN,
                 MODE, DEFAULT_VALUE, SOURCE_COLUMN, STRATEGIES_DEFAULT)


class DefaultHandlerModel:
    def __init__(self, model, config):
        """
        Initialize the handler for the model.
        :param model: The PandasTable Model containing the Data.
        :param config: Dictionary containing strategies and settings for each column.
        """
        self._model = model
        # TODO: Check what happens with non inplace operations
        self.model = model._data_frame
        self.config = config
        self.model_index = self.model.index.name

    def get_default(self, column_name, row_index=None):
        """
        Get the default value for a column based on its strategy.
        :param column_name: The name of the column to compute the default for.
        :param row_index: Optional index of the row (needed for some strategies).
        :return: The computed default value.
        """
        source_column = column_name
        if column_name not in self.config:
            if "default_config" in self.config:
                column_name = "default_config"
            else:
                return ""

        column_config = self.config[column_name]
        strategy = column_config.get("strategy", NO_DEFAULT)
        default_value = column_config.get(DEFAULT_VALUE, "")

        if strategy == USE_DEFAULT:
            return default_value
        elif strategy == NO_DEFAULT:
            return ""
        elif strategy == MIN_COLUMN:
            return self._min_column(column_name)
        elif strategy == MAX_COLUMN:
            return self._max_column(column_name)
        elif strategy == COPY_FROM:
            return self._copy_column(column_name, column_config, row_index)
        elif strategy == MODE:
            column_config[SOURCE_COLUMN] = source_column
            return self._majority_vote(column_name, column_config)
        else:
            raise ValueError(f"Unknown strategy '{strategy}' for column '{column_name}'.")

    def _min_column(self, column_name):
        if column_name in self.model:
            column_data = self.model[column_name].replace("", np.nan).dropna()
            if not column_data.empty:
                return column_data.min()
        return ""

    def _max_column(self, column_name):
        if column_name in self.model:
            column_data = self.model[column_name].replace("", np.nan).dropna()
            if not column_data.empty:
                return column_data.max()
        return ""

    def _copy_column(self, column_name, config, row_index):
        source_column = config.get(SOURCE_COLUMN, column_name)
        source_column_valid = (
            source_column in self.model or source_column == self.model_index
        )
        if source_column and source_column_valid and row_index is not None:
            prefix = config.get("prefix", "")
            if row_index in self.model.index:
                if source_column == self.model_index:
                    return f"{prefix}{row_index}"
                value = f"{prefix}{self.model.at[row_index, source_column]}"
                return value if pd.notna(value) else ""
        return ""

    def _majority_vote(self, column_name, config):
        """Use the most frequent value in the column as the default.

        Defaults to last used value in case of a tie.
        """
        source_column = config.get(SOURCE_COLUMN, column_name)
        source_column_valid = (
            source_column in self.model or source_column == self.model_index
        )
        if source_column and source_column_valid:
            valid_values = copy.deepcopy(self.model[source_column][:-1])
            valid_values = valid_values.iloc[::-1]
            if valid_values.empty:
                return ""
            value_counts = Counter(valid_values)
            return value_counts.most_common(1)[0][0]
        return ""



"""
Models
======

This package contains the models for the PEtab edit GUI.
"""

from .pandas_table_model import MeasurementModel, ObservableModel, \
    ParameterModel, ConditionModel, PandasTableModel
from .sbml_model import SbmlViewerModel
from .petab_model import PEtabModel


"""Classes for the controllers of the tables in the GUI."""
from PySide6.QtWidgets import QInputDialog, QMessageBox, QFileDialog, \
    QCompleter, QAbstractItemView
import numpy as np
import pandas as pd
import petab.v1 as petab
from PySide6.QtCore import Signal, QObject, QModelIndex, Qt, QTimer
from pathlib import Path
from ..models.pandas_table_model import PandasTableModel, \
    PandasTableFilterProxy
from ..settings_manager import settings_manager
from ..views.table_view import TableViewer, SingleSuggestionDelegate, \
    ColumnSuggestionDelegate, ComboBoxDelegate, ParameterIdSuggestionDelegate
from ..utils import get_selected, process_file
from .utils import prompt_overwrite_or_append
from ..C import COLUMN
import re


class TableController(QObject):
    """Base class for table controllers."""
    overwritten_df = Signal()  # Signal to mother controller

    def __init__(
        self,
        view: TableViewer,
        model: PandasTableModel,
        logger,
        mother_controller
    ):
        """Initialize the table controller.

        Parameters
        ----------
        view: TableViewer
            The view of the table.
        model: PandasTableModel
            The model of the table.
        logger:
            Handles all logging tasks
        mother_controller: MainController
            The main controller of the application. Needed for signal
            forwarding.
        """
        super().__init__()
        self.view = view
        self.model = model
        self.model.view = self.view.table_view
        self.proxy_model = PandasTableFilterProxy(model)
        self.logger = logger
        self.check_petab_lint_mode = True
        self.mother_controller = mother_controller
        self.view.table_view.setModel(self.proxy_model)
        self.setup_connections()
        self.setup_connections_specific()

        self.completers = {}
        self.setup_completers()

    def setup_completers(self):
        pass

    def release_completers(self):
        """Sets the completers to None. Safety Measure."""
        if not self.completers:
            return
        for column_index in range(self.model.columnCount()):
            self.view.table_view.setItemDelegateForColumn(column_index, None)
        self.completers = {}

    def setup_connections_specific(self):
        """Will be implemented in child controllers."""
        pass

    def setup_connections(self):
        """Setup connections to the view.

        Only handles connections from within the table controllers.
        """
        self.model.new_log_message.connect(
            self.logger.log_message
        )
        self.model.cell_needs_validation.connect(
            self.validate_changed_cell
        )
        self.model.inserted_row.connect(
            self.set_index_on_new_row
        )
        self.model.fill_defaults.connect(
            self.model.get_default_values, Qt.QueuedConnection
        )
        settings_manager.settings_changed.connect(
            self.update_defaults
        )

    def setup_context_menu(self, actions):
        """Setup context menu for this table."""
        view = self.view.table_view
        view.setup_context_menu(actions)

    def validate_changed_cell(self, row, column):
        """Validate the changed cell and whether its linting is correct."""
        if not self.check_petab_lint_mode:
            return
        row_data = self.model.get_df().iloc[row]
        index_name = self.model.get_df().index.name
        row_data = row_data.to_frame().T
        row_data.index.name = index_name
        try:
            self.check_petab_lint(row_data)
            for col in range(self.model.columnCount()):
                self.model.discard_invalid_cell(row, col)
            error_message = None
        except Exception as e:
            error_message = str(e)
            self.logger.log_message(
                f"PEtab linter failed at row {row}, column {column}: "
                f"{error_message}",
                color="red"
            )
        # Update invalid cells based on the error state
        if error_message:
            self.model.add_invalid_cell(row, column)
        else:
            self.model.discard_invalid_cell(row, column)
        self.model.notify_data_color_change(row, column)

    def open_table(self, file_path=None, separator=None, mode="overwrite"):
        if not file_path:
            # Open a file dialog to select the CSV or TSV file
            file_path, _ = QFileDialog.getOpenFileName(
                self.view, "Open CSV or TSV", "",
                "CSV/TSV/TXT Files (*.csv *.tsv *.txt)"
            )
        # just in case anything goes wrong here
        if not file_path:
            return
        # convert the file path to a Path object if it is a string
        if type(file_path) is str:
            file_path = Path(file_path)

        if separator is None:
            actionable, separator = process_file(file_path, self.logger)
            if actionable in ["yaml", "sbml", "data_matrix", None]:  # no table
                return
        try:
            if self.model.table_type == "measurement":
                new_df = pd.read_csv(file_path, sep=separator)
            else:
                new_df = pd.read_csv(file_path, sep=separator, index_col=0)
        except Exception as e:
            self.view.log_message(
                f"Failed to read file: {str(e)}",
                color="red"
            )
            return
        dtypes = {
            col: self.model._allowed_columns.get(
                col, {"type": np.object_}
            )["type"] for col in new_df.columns
        }
        new_df = new_df.astype(dtypes)
        if mode is None:
            mode = prompt_overwrite_or_append(self)
        # Overwrite or append the table with the new DataFrame
        if mode == "append":
            self.append_df(new_df)
        elif mode == "overwrite":
            self.overwrite_df(new_df)
            self.model.reset_invalid_cells()

    def overwrite_df(self, new_df: pd.DataFrame):
        # TODO: Mother controller connects to overwritten_df signal. Set df
        #  in petabProblem and unsaved changes to True
        """Overwrite the DataFrame of the model with the data from the view."""
        self.proxy_model.setSourceModel(None)
        self.model.beginResetModel()
        self.model._data_frame = new_df
        self.model.endResetModel()
        self.logger.log_message(
            f"Overwrote the {self.model.table_type} table with new data.",
            color="green"
        )
        # test: overwrite the new model as source model
        self.proxy_model.setSourceModel(self.model)
        # change default sizing
        self.view.table_view.reset_column_sizes()
        self.overwritten_df.emit()

    def append_df(self, new_df: pd.DataFrame):
        """Append the DataFrame of the model with the data from the view.

        Merges two DataFrames:
            1. Columns are the union of both DataFrame columns.
            2. Rows are the union of both DataFrame rows (duplicates removed)
        """
        self.model.beginResetModel()
        combined_df = pd.concat(
            [self.model.get_df(), new_df], axis=0
        )
        combined_df = combined_df[~combined_df.index.duplicated(keep="first")]
        self.model._data_frame = combined_df
        self.proxy_model.setSourceModel(None)
        self.proxy_model.setSourceModel(self.model)
        self.model.endResetModel()
        self.logger.log_message(
            f"Appended the {self.model.table_type} table with new data.",
            color="green"
        )
        # test: overwrite the new model as source model
        self.overwritten_df.emit()

    def clear_table(self):
        """Clear the table."""
        self.model.clear_table()

    def delete_row(self):
        """Delete the selected row(s) from the table."""
        table_view = self.view.table_view

        selected_rows = get_selected(table_view)
        if not selected_rows:
            return
        self.model.update_invalid_cells(selected_rows, mode="rows")
        for row in sorted(selected_rows, reverse=True):
            if row >= self.model.rowCount() - 1:
                continue
            self.model.delete_row(row)
            self.logger.log_message(
                f"Deleted row {row} from {self.model.table_type} table."
                f" Data: {self.model.get_df().iloc[row].to_dict()}",
                color="orange"
            )
        self.model.something_changed.emit(True)

    def add_row(self):
        """Add a row to the datatable"""
        row_count = self.model.rowCount() - 1
        if self.model.insertRows(row_count, 1):
            new_row_index = self.model.index(row_count, 0)

            selection_model = self.view.table_view.selectionModel()
            if selection_model:
                selection_model.select(
                    new_row_index,
                    selection_model.SelectionFlag.ClearAndSelect
                )
            self.view.table_view.scrollTo(new_row_index)
            self.view.table_view.setCurrentIndex(new_row_index)

    def delete_column(self):
        """Delete the selected column(s) from the table."""
        table_view = self.view.table_view

        selected_columns = get_selected(table_view, mode=COLUMN)
        if not selected_columns:
            return
        deleted_columns = set()
        for column in sorted(selected_columns, reverse=True):
            # safely delete potential item delegates
            allow_del, column_name = self.model.allow_column_deletion(column)
            if not allow_del:
                self.logger.log_message(
                    f"Cannot delete column {column_name}, as it is a "
                    f"required column!",
                    color = "red"
                )
                continue
            if column_name in self.completers:
                self.view.table_view.setItemDelegateForColumn(column, None)
                del self.completers[column_name]
            self.model.delete_column(column)
            self.logger.log_message(
                f"Deleted column '{column_name}' from {self.model.table_type} table.",
                color="orange"
            )
            deleted_columns.add(column)
        self.model.update_invalid_cells(deleted_columns, mode="columns")
        self.model.something_changed.emit(True)

    def add_column(self, column_name: str = None):
        """Add a column to the datatable"""
        if not column_name:
            column_name, ok = QInputDialog.getText(
                self.view,
                "Add Column",
                "Enter the name of the new column:"
            )
            if not ok:
                return
        self.model.insertColumn(column_name)

    def set_index_on_new_row(self, index: QModelIndex):
        """Set the index of the model when a new row is added."""
        self.view.table_view.setCurrentIndex(index)

    def filter_table(self, text):
        """Filter the table."""
        self.proxy_model.setFilterRegularExpression(text)
        self.proxy_model.setFilterKeyColumn(-1)

    def remove_filter(self):
        """Remove the filter from the table."""
        self.proxy_model.setFilterRegularExpression("")
        self.proxy_model.setFilterKeyColumn(-1)

    def copy_to_clipboard(self):
        """Copy the currently selected cells to the clipboard."""
        self.view.copy_to_clipboard()

    def paste_from_clipboard(self):
        """Paste the clipboard content to the currently selected cells."""
        self.check_petab_lint_mode = False
        self.view.paste_from_clipboard()
        self.check_petab_lint_mode = True
        try:
            self.check_petab_lint()
        except Exception as e:
            self.logger.log_message(
                f"PEtab linter failed after copying: {str(e)}",
                color="red"
            )

    def check_petab_lint(self, row_data):
        """Check a single row of the model with petablint."""
        raise NotImplementedError(
            "This method must be implemented in child classes."
        )

    def find_text(
        self, text, case_sensitive=False, regex=False, whole_cell=False
    ):
        """Efficiently find all matching cells."""
        df = self.model.get_df()

        # Search in the main DataFrame
        if regex:
            pattern = re.compile(text, 0 if case_sensitive else re.IGNORECASE)
            mask = df.map(lambda cell: bool(pattern.fullmatch(str(cell))) if whole_cell else bool(pattern.search(str(cell))))
        else:
            text = text.lower() if not case_sensitive else text
            mask = df.map(lambda cell: text == str(cell).lower() if whole_cell else text in str(cell).lower()) if not case_sensitive else \
                   df.map(lambda cell: text == str(cell) if whole_cell else text in str(cell))

        # Find matches
        match_indices = list(zip(*mask.to_numpy().nonzero()))
        table_matches = [(row, col + self.model.column_offset) for row, col in match_indices]

        # Search in the index if it's named
        index_matches = []
        if isinstance(df.index, pd.Index) and df.index.name:
            if regex:
                index_mask = df.index.to_series().map(lambda idx: bool(pattern.fullmatch(str(idx))) if whole_cell else bool(pattern.search(str(idx))))
            else:
                index_mask = df.index.to_series().map(lambda idx: text == str(idx).lower() if whole_cell else text in str(idx).lower()) if not case_sensitive else \
                             df.index.to_series().map(lambda idx: text == str(idx) if whole_cell else text in str(idx))

            index_matches = [(df.index.get_loc(idx), 0) for idx in index_mask[index_mask].index]

        all_matches = index_matches + table_matches

        # 🔹 Highlight matched text
        self.highlight_text(all_matches)
        return all_matches

    def highlight_text(self, matches):
        """Color the text of all matched cells in yellow."""
        self.model.highlighted_cells = set(matches)
        top_left = self.model.index(0, 0)
        bottom_right = self.model.index(self.model.rowCount() - 1,
                                        self.model.columnCount() - 1)
        self.model.dataChanged.emit(top_left, bottom_right,
                                    [Qt.ForegroundRole])

    def cleanse_highlighted_cells(self):
        """Cleanses the highlighted cells."""
        self.model.highlighted_cells = set()
        top_left = self.model.index(0, 0)
        bottom_right = self.model.index(self.model.rowCount() - 1,
                                        self.model.columnCount() - 1)
        self.model.dataChanged.emit(top_left, bottom_right,
                                    [Qt.ForegroundRole])

    def focus_match(self, match, with_focus: bool = False):
        """Focus and select the given match in the table."""
        if match is None:
            self.view.table_view.clearSelection()
            return
        row, col = match
        index = self.model.index(row, col)
        if not index.isValid():
            return
        proxy_index = self.view.table_view.model().mapFromSource(index)
        if not proxy_index.isValid():
            return

        self.view.table_view.setCurrentIndex(proxy_index)
        self.view.table_view.scrollTo(
                proxy_index, QAbstractItemView.EnsureVisible
        )
        if with_focus:
            self.view.table_view.setFocus()

    def replace_text(self, row, col, replace_text, search_text, case_sensitive, regex):
        """Replace the text in the given cell and update highlights."""
        index = self.model.index(row, col)
        original_text = self.model.data(index, Qt.DisplayRole)

        if not original_text:
            return

        if regex:
            pattern = re.compile(search_text, 0 if case_sensitive else re.IGNORECASE)
            new_text = pattern.sub(replace_text, original_text)
        else:
            if not case_sensitive:
                search_text = re.escape(search_text.lower())
                new_text = re.sub(search_text, replace_text, original_text, flags=re.IGNORECASE)
            else:
                new_text = original_text.replace(search_text, replace_text)

        if new_text != original_text:
            self.model.setData(index, new_text, Qt.EditRole)
            self.model.highlighted_cells.discard((row, col))
            self.model.dataChanged.emit(index, index, [Qt.DisplayRole])

    def replace_all(
        self, search_text, replace_text, case_sensitive=False, regex=False
    ):
        """Replace all occurrences of the search term in the Model."""
        if not search_text or not replace_text:
            return

        df = self.model._data_frame
        if regex:
            pattern = re.compile(search_text,
                                 0 if case_sensitive else re.IGNORECASE)
            df.replace(to_replace=pattern, value=replace_text, regex=True,
                       inplace=True)
        else:
            if not case_sensitive:
                df.replace(
                    to_replace=re.escape(search_text),
                    value=replace_text,
                    regex=True,
                    inplace=True
                )
            else:
                df.replace(to_replace=search_text, value=replace_text,
                           inplace=True)

        # Replace in the index as well
        if isinstance(df.index, pd.Index) and df.index.name:
            index_map = {
                idx: pattern.sub(replace_text, str(idx)) if regex else str(
                    idx).replace(search_text, replace_text)
                for idx in df.index if search_text in str(idx)}
            if index_map:
                df.rename(index=index_map, inplace=True)

    def get_columns(self):
        """Get the columns of the table."""
        df = self.model.get_df()
        # if it is a named index, add it to the columns
        if df.index.name:
            return [df.index.name] + df.columns.tolist()
        return df.columns.tolist()

    def update_defaults(self, settings_changed):
        """Update the default values of the model."""
        # if the signal is not "table_defaults/table_name" return
        if not settings_changed.startswith("table_defaults"):
            return
        table_name = settings_changed.split("/")[1]
        if table_name != self.model.table_type:
            return
        self.model.default_handler.config = (
            settings_manager.get_table_defaults(self.model.table_type)
        )


class MeasurementController(TableController):
    """Controller of the Measurement table."""

    def check_petab_lint(self, row_data: pd.DataFrame = None):
        """Check a number of rows of the model with petablint."""
        if row_data is None:
            row_data = self.model.get_df()
        # Can this be done more elegantly?
        observable_df = self.mother_controller.model.observable.get_df()
        return petab.check_measurement_df(
            row_data,
            observable_df=observable_df,
        )

    def rename_value(self, old_id: str, new_id: str, column_names: str | list[str]):
        """Rename the observables in the measurement_df.

        Triggered by changes in the original observable_df id.

        Parameters
        ----------
        old_id:
            The old observable_id, which was changed.
        new_id:
            The new observable_id.
        """
        if not isinstance(column_names, list):
            column_names = [column_names]

        # Find occurences
        mask = self.model._data_frame[column_names].eq(old_id)
        if mask.any().any():
            self.model._data_frame.loc[mask] = new_id
            changed_rows = mask.any(axis=1)
            first_row, last_row = (
                changed_rows.idxmax(), changed_rows[::-1].idxmax()
            )
            top_left = self.model.index(first_row, 1)
            bottom_right = self.model.index(
                last_row, self.model.columnCount() - 1
            )
            self.model.dataChanged.emit(
                top_left, bottom_right, [Qt.DisplayRole, Qt.EditRole]
            )

            # Emit change signal
            self.model.something_changed.emit(True)

    def copy_noise_parameters(
        self,
        observable_id: str,
        condition_id: str | None = None
    ) -> str:
        """Copies noise parameter from measurements already in the table.

        Measurements of similar observables are most likely assumed to
        share a noise model. Therefore, noise parameters are copied. Prefers
        matching condition_id to copy. If not Matching condition_id,
        will copy from any matching row.

        Parameters:
        ----------
        observable_id:
            The observable_id of the new measurement.
        condition_id:
            The condition_id of the new measurement.

        Returns:
            The noise parameter that has been copied, or "" if no noise
            parameter could be copied.
        """
        measurement_df = self.model.measurement._data_frame
        matching_rows = measurement_df[
            measurement_df["observableId"] == observable_id]
        if matching_rows.empty:
            return ""
        if not condition_id:
            return matching_rows["noiseParameters"].iloc[0]
        preferred_row = matching_rows[
            matching_rows["simulationConditionId"] == condition_id]
        if not preferred_row.empty:
            return preferred_row["noiseParameters"].iloc[0]
        else:
            return matching_rows["noiseParameters"].iloc[0]

    def upload_data_matrix(self):
        """Upload a data matrix to the measurement table.

        Opens a FileDialog to select a CSV file with the data matrix.
        The data matrix is a CSV file with the following columns:
        - time: Either "Time", "time" or "t". Time points of the measurements.
        - observable_ids: Observables measured at the given timepoints.
        """
        file_name, _ = QFileDialog.getOpenFileName(
            self.view,
            "Open Data Matrix",
            "",
            "CSV Files (*.csv);;TSV Files (*.tsv)"
        )
        if file_name:
            self.process_data_matrix_file(file_name, "append")

    def process_data_matrix_file(self, file_name, mode, separator=None):
        """Process the data matrix file.

        Upload the data matrix. Then populate the measurement table with the
        new measurements. Additionally, triggers checks for observable_ids.
        """
        try:
            data_matrix = self.load_data_matrix(file_name, separator)
            if data_matrix is None or data_matrix.empty:
                return

            cond_dialog = ConditionInputDialog()
            if cond_dialog.exec():
                conditions = cond_dialog.get_condition_id()
                condition_id = conditions.get("conditionId", "")
                preeq_id = conditions.get("preeq_id", "")
            if mode == "overwrite":
                self.model.clear_table()
            self.populate_tables_from_data_matrix(
                data_matrix, condition_id, preeq_id
            )
            self.model.something_changed.emit(True)

        except Exception as e:
            self.logger.log_message(
                f"An error occurred while uploading the data matrix: {str(e)}",
                color="red"
            )

    def load_data_matrix(self, file_name, separator=None):
        """Loads in the data matrix. Checks for the 'time' column."""
        data_matrix = pd.read_csv(
            file_name, delimiter=separator
        )
        if not any(
            col in data_matrix.columns for col in ["Time", "time", "t"]):
            self.logger.log_message(
                "Invalid File, the file must contain a 'Time' column. "
                "Please ensure that the file contains a 'Time'",
                color="red"
            )
            return None

        time_column = next(
            col for col in ["Time", "time", "t"] if col in data_matrix.columns
        )
        return data_matrix.rename(columns={time_column: "time"})

    def populate_tables_from_data_matrix(
        self, data_matrix, condition_id, preeq_id: str = ""
    ):
        """Populate the measurement table from the data matrix."""
        for col in data_matrix.columns:
            if col == "time":
                continue
            observable_id = col
            self.model.relevant_id_changed.emit(
                observable_id, "", "observable"
            )
            self.model.relevant_id_changed.emit(condition_id, "", "condition")
            if preeq_id:
                self.model.relevant_id_changed.emit(preeq_id, "", "condition")
            self.add_measurement_rows(
                data_matrix[["time", observable_id]],
                observable_id,
                condition_id,
                preeq_id
            )

    def add_measurement_rows(
        self, data_matrix,
        observable_id,
        condition_id: str = "",
        preeq_id: str = ""
    ):
        """Adds multiple rows to the measurement table."""
        # check number of rows and signal row insertion
        rows = data_matrix.shape[0]
        # get current number of rows
        current_rows = self.model.get_df().shape[0]
        self.model.insertRows(position=None,
                              rows=rows)  # Fills the table with empty rows
        top_left = self.model.createIndex(current_rows, 0)
        for i_row, (_, row) in enumerate(data_matrix.iterrows()):
            self.model.fill_row(
                i_row + current_rows,
                data={
                    "observableId": observable_id,
                    "time": row["time"],
                    "measurement": row[observable_id],
                    "simulationConditionId": condition_id,
                    "preequilibrationConditionId": preeq_id
                }
            )
        bottom, right = (x - 1 for x in self.model.get_df().shape)
        bottom_right = self.model.createIndex(bottom, right)
        self.model.dataChanged.emit(top_left, bottom_right)
        self.logger.log_message(
            f"Added {rows} measurements to the measurement table.",
            color="green"
        )

    def setup_completers(self):
        """Set completers for the measurement table."""
        table_view = self.view.table_view
        # observableId
        observableId_index = self.model.return_column_index("observableId")
        if observableId_index > -1:
            self.completers["observableId"] = ColumnSuggestionDelegate(
                self.mother_controller.model.observable, "observableId"
            )
            table_view.setItemDelegateForColumn(
                observableId_index,
                self.completers["observableId"]
            )
        # preequilibrationConditionId
        preequilibrationConditionId_index = self.model.return_column_index(
            "preequilibrationConditionId"
        )
        if preequilibrationConditionId_index > -1:
            self.completers[
                "preequilibrationConditionId"] = ColumnSuggestionDelegate(
                self.mother_controller.model.condition, "conditionId"
            )
            table_view.setItemDelegateForColumn(
                preequilibrationConditionId_index,
                self.completers["preequilibrationConditionId"]
            )
        # simulationConditionId
        simulationConditionId_index = self.model.return_column_index(
            "simulationConditionId"
        )
        if simulationConditionId_index > -1:
            self.completers[
                "simulationConditionId"] = ColumnSuggestionDelegate(
                self.mother_controller.model.condition, "conditionId"
            )
            table_view.setItemDelegateForColumn(
                simulationConditionId_index,
                self.completers["simulationConditionId"]
            )
        # noiseParameters
        noiseParameters_index = self.model.return_column_index(
            "noiseParameters")
        if noiseParameters_index > -1:
            self.completers["noiseParameters"] = SingleSuggestionDelegate(
                self.model, "observableId", afix="sd_"
            )
            table_view.setItemDelegateForColumn(
                noiseParameters_index,
                self.completers["noiseParameters"]
            )


class ConditionController(TableController):
    """Controller of the Condition table."""
    condition_2be_renamed = Signal(str, str)  # Signal to mother controller

    def update_handler_model(self):
        """Update the handler model."""
        self.model.default_handler.model = self.model._data_frame

    def setup_connections_specific(self):
        """Setup connections specific to the condition controller.

        Only handles connections from within the table controllers.
        """
        self.model.relevant_id_changed.connect(
            self.maybe_rename_condition
        )
        self.overwritten_df.connect(
            self.update_handler_model
        )

    def check_petab_lint(self, row_data: pd.DataFrame = None):
        """Check a number of rows of the model with petablint."""
        if row_data is None:
            row_data = self.model.get_df()
        observable_df = self.mother_controller.model.observable.get_df()
        sbml_model = self.mother_controller.model.sbml.get_current_sbml_model()
        return petab.check_condition_df(
            row_data,
            observable_df=observable_df,
            model=sbml_model,
        )

    def maybe_rename_condition(self, new_id, old_id):
        """Potentially rename condition_ids in measurement_df.

        Opens a dialog to ask the user if they want to rename the conditions.
        If so, emits a signal to rename the conditions in the measurement_df.
        """
        if old_id not in self.mother_controller.measurement_controller.model.get_df()["simulationConditionId"].values:
            return
        reply = QMessageBox.question(
            self.view, 'Rename Condition',
            f'Do you want to rename condition "{old_id}" to "{new_id}" '
            f'in all measurements?',
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.No
        )
        if reply == QMessageBox.Yes:
            self.logger.log_message(
                f"Renaming condition '{old_id}' to '{new_id}' in all "
                f"measurements",
                color="green"
            )
            self.condition_2be_renamed.emit(old_id, new_id)


    def maybe_add_condition(self, condition_id, old_id=None):
        """Add a condition to the condition table if it does not exist yet."""
        if condition_id in self.model.get_df().index or not condition_id:
            return
        # add a row
        self.model.insertRows(position=None, rows=1)
        self.model.fill_row(
            self.model.get_df().shape[0] - 1,
            data={"conditionId": condition_id}
        )
        self.model.cell_needs_validation.emit(
            self.model.get_df().shape[0] - 1, 0
        )
        self.logger.log_message(
            f"Automatically added condition '{condition_id}' to the condition "
            f"table.",
            color="green"
        )

    def setup_completers(self):
        """Set completers for the condition table."""
        table_view = self.view.table_view
        # conditionName
        conditionName_index = self.model.return_column_index("conditionName")
        if conditionName_index > -1:
            self.completers["conditionName"] = SingleSuggestionDelegate(
                self.model, "conditionId")
            table_view.setItemDelegateForColumn(
                conditionName_index,
                self.completers["conditionName"]
            )
        for column in self.model.get_df().columns:
            if column in ["conditionId", "conditionName"]:
                continue
            column_index = self.model.return_column_index(column)
            if column_index > -1:
                self.completers[column] = ColumnSuggestionDelegate(
                    self.model, column, QCompleter.PopupCompletion
                )
                table_view.setItemDelegateForColumn(
                    column_index,
                    self.completers[column]
                )


class ObservableController(TableController):
    """Controller of the Observable table."""
    observable_2be_renamed = Signal(str, str)  # Signal to mother controller

    def update_handler_model(self):
        """Update the handler model."""
        self.model.default_handler.model = self.model._data_frame

    def setup_completers(self):
        """Set completers for the observable table."""
        table_view = self.view.table_view
        # observableName
        observableName_index = self.model.return_column_index("observableName")
        if observableName_index > -1:
            self.completers["observableName"] = SingleSuggestionDelegate(
                self.model, "observableId")
            table_view.setItemDelegateForColumn(
                observableName_index,
                self.completers["observableName"]
            )
        # observableTransformation
        observableTransformation_index = self.model.return_column_index(
            "observableTransformation"
        )
        if observableTransformation_index > -1:
            self.completers["observableTransformation"] = ComboBoxDelegate(
                ["lin", "log", "log10"]
            )
            table_view.setItemDelegateForColumn(
                observableTransformation_index,
                self.completers["observableTransformation"]
            )
        # noiseFormula
        noiseFormula_index = self.model.return_column_index("noiseFormula")
        if noiseFormula_index > -1:
            self.completers["noiseFormula"] = SingleSuggestionDelegate(
                self.model, "observableId", afix="noiseParameter1_"
            )
            table_view.setItemDelegateForColumn(
                noiseFormula_index,
                self.completers["noiseFormula"]
            )
        # noiseDistribution
        noiseDistribution_index = self.model.return_column_index(
            "noiseDistribution"
        )
        if noiseDistribution_index > -1:
            self.completers["noiseDistribution"] = ComboBoxDelegate(
                ["normal", "laplace"]
            )
            table_view.setItemDelegateForColumn(
                noiseDistribution_index,
                self.completers["noiseDistribution"]
            )

    def setup_connections_specific(self):
        """Setup connections specific to the observable controller.

        Only handles connections from within the table controllers.
        """
        self.model.relevant_id_changed.connect(
            self.maybe_rename_observable
        )
        self.overwritten_df.connect(
            self.update_handler_model
        )

    def check_petab_lint(self, row_data: pd.DataFrame = None):
        """Check a number of rows of the model with petablint."""
        if row_data is None:
            row_data = self.model.get_df()
        return petab.check_observable_df(row_data)

    def maybe_rename_observable(self, new_id, old_id):
        """Potentially rename observable_ids in measurement_df.

        Opens a dialog to ask the user if they want to rename the observables.
        If so, emits a signal to rename the observables in the measurement_df.
        """
        if old_id not in self.mother_controller.measurement_controller.model.get_df()["observableId"].values:
            return
        reply = QMessageBox.question(
            self.view, 'Rename Observable',
            f'Do you want to rename observable "{old_id}" to "{new_id}" '
            f'in all measurements?',
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.No
        )
        if reply == QMessageBox.Yes:
            self.logger.log_message(
                f"Renaming observable '{old_id}' to '{new_id}' in all "
                f"measurements",
                color="green"
            )
            # TODO: connect this signal with the measurement function
            self.observable_2be_renamed.emit(old_id, new_id)

    def maybe_add_observable(self, observable_id, old_id=None):
        """Add an observable to the observable table if it does not exist yet.

        Currently, `old_id` is not used.
        """
        if observable_id in self.model.get_df().index or not observable_id:
            return
        # add a row
        self.model.insertRows(position=None, rows=1)
        self.model.fill_row(
            self.model.get_df().shape[0] - 1,
            data={"observableId": observable_id}
        )
        self.model.cell_needs_validation.emit(
            self.model.get_df().shape[0] - 1, 0
        )
        self.logger.log_message(
            f"Automatically added observable '{observable_id}' to the "
            f"observable table.",
            color="green"
        )


class ParameterController(TableController):
    """Controller of the Parameter table."""

    def setup_connections_specific(self):
        """Connect signals specific to the parameter controller."""
        self.overwritten_df.connect(
            self.update_handler_model
        )

    def update_handler_model(self):
        """Update the handler model."""
        self.model.default_handler.model = self.model._data_frame

    def setup_completers(self):
        """Set completers for the parameter table."""
        table_view = self.view.table_view
        # parameterName
        parameterName_index = self.model.return_column_index("parameterName")
        if parameterName_index > -1:
            self.completers["parameterName"] = SingleSuggestionDelegate(
                self.model, "parameterId")
            table_view.setItemDelegateForColumn(
                parameterName_index,
                self.completers["parameterName"]
            )
        # parameterScale
        parameterScale_index = self.model.return_column_index("parameterScale")
        if parameterScale_index > -1:
            self.completers["parameterScale"] = ComboBoxDelegate(
                ["lin", "log", "log10"]
            )
            table_view.setItemDelegateForColumn(
                parameterScale_index,
                self.completers["parameterScale"]
            )
        # lowerBound
        lowerBound_index = self.model.return_column_index("lowerBound")
        if lowerBound_index > -1:
            self.completers["lowerBound"] = ColumnSuggestionDelegate(
                self.model, "lowerBound", QCompleter.PopupCompletion
            )
            table_view.setItemDelegateForColumn(
                lowerBound_index,
                self.completers["lowerBound"]
            )
        # upperBound
        upperBound_index = self.model.return_column_index("upperBound")
        if upperBound_index > -1:
            self.completers["upperBound"] = ColumnSuggestionDelegate(
                self.model, "upperBound", QCompleter.PopupCompletion
            )
            table_view.setItemDelegateForColumn(
                upperBound_index,
                self.completers["upperBound"]
            )
        # estimate
        estimate_index = self.model.return_column_index("estimate")
        if estimate_index > -1:
            self.completers["estimate"] = ComboBoxDelegate(
                ["1", "0"]
            )
            table_view.setItemDelegateForColumn(
                estimate_index,
                self.completers["estimate"]
            )
        # parameterId: retrieved from the sbml model
        parameterId_index = self.model.return_column_index("parameterId")
        sbml_model = self.mother_controller.model.sbml
        if parameterId_index > -1:
            self.completers["parameterId"] = ParameterIdSuggestionDelegate(
                par_model=self.model,
                sbml_model=sbml_model
            )
            table_view.setItemDelegateForColumn(
                parameterId_index,
                self.completers["parameterId"]
            )

    def check_petab_lint(self, row_data: pd.DataFrame = None):
        """Check a number of rows of the model with petablint."""
        if row_data is None:
            row_data = self.model.get_df()
        observable_df = self.mother_controller.model.observable.get_df()
        measurement_df = self.mother_controller.model.measurement.get_df()
        condition_df = self.mother_controller.model.condition.get_df()
        sbml_model = self.mother_controller.model.sbml.get_current_sbml_model()
        return petab.check_parameter_df(
            row_data,
            observable_df=observable_df,
            measurement_df=measurement_df,
            condition_df=condition_df,
            model=sbml_model,
        )
